from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Callable
import math
from fontTools.misc import transform
from fontParts.base.base import (
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    PointPositionMixin,
    IdentifierMixin,
    dynamicProperty,
    reference,
)
from fontParts.base import normalizers
from fontParts.base.compatibility import GuidelineCompatibilityReporter
from fontParts.base.color import Color
from fontParts.base.deprecated import DeprecatedGuideline, RemovedGuideline
from fontParts.base.annotations import (
    AffineTransformationLike,
    RGBALike,
    RGBA,
    IntFloatType,
)

if TYPE_CHECKING:
    from fontParts.base.font import BaseFont
    from fontParts.base.layer import BaseLayer
    from fontParts.base.glyph import BaseGlyph


class BaseGuideline(
    BaseObject,
    TransformationMixin,
    DeprecatedGuideline,
    RemovedGuideline,
    PointPositionMixin,
    InterpolationMixin,
    IdentifierMixin,
    SelectionMixin,
):
    """Represent the basis for a guideline object.

    This object is almost always created with :meth:`BaseGlyph.appendGuideline`.
    An orphan guideline can be created like this::

        >>> guideline = RGuideline()

    """

    copyAttributes: tuple[str, ...] = ("x", "y", "angle", "name", "color")

    def _reprContents(self) -> list[str]:
        contents = []
        if self.name is not None:
            contents.append(f"'{self.name}'")
        if self.layer is not None:
            contents.append(f"('{self.layer.name}')")
        return contents

    # -------
    # Parents
    # -------

    # Glyph

    _glyph: Callable[[], BaseGlyph] | None = None

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get or set the guideline's parent glyph object.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the guideline
            or :obj:`None`.
        :raises AssertionError: If attempting to set the glyph when it
            has already been set.

        Example::

            >>> glyph = guideline.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph: BaseGlyph | Callable[[], BaseGlyph] | None) -> None:
        if self._font is not None:
            raise AssertionError("font for guideline already set")
        if self._glyph is not None:
            raise AssertionError("glyph for guideline already set")
        if glyph is not None:
            glyph = reference(glyph)
        self._glyph = glyph

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the guideline's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the guideline
            or :obj:`None`.

        Example::

            >>> layer = guideline.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._glyph is None:
            return None
        return self.glyph.layer

    # Font

    _font: Callable[[], BaseFont] | None = None

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the guideline's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the guideline
            or :obj:`None`.

        Example::

            >>> font = guideline.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._font is not None:
            return self._font()
        elif self._glyph is not None:
            return self.glyph.font
        return None

    def _set_font(self, font: BaseFont | Callable[[], BaseFont] | None) -> None:
        if self._font is not None:
            raise AssertionError("font for guideline already set")
        if self._glyph is not None:
            raise AssertionError("glyph for guideline already set")
        if font is not None:
            font = reference(font)
        self._font = font

    # --------
    # Position
    # --------

    # x

    x: dynamicProperty = dynamicProperty(
        "base_x",
        """Get or set the guideline's x-coordinate.

        The value must be an :class:`int` or a :class:`flat`.

        :return: An :class:`int` or a :class:`flat` representing the
            x-coordinate of the guideline.

        Example::

            >>> guideline.x
            100
            >>> guideline.x = 101

        """,
    )

    def _get_base_x(self) -> IntFloatType:
        value = self._get_x()
        if value is None:
            return 0
        value = normalizers.normalizeX(value)
        return value

    def _set_base_x(self, value: IntFloatType) -> None:
        if value is None:
            value = 0
        else:
            value = normalizers.normalizeX(value)
        self._set_x(value)

    def _get_x(self) -> IntFloatType:
        """Get the native guideline's x-coordinate.

        This is the environment implementation of the :attr:`BaseGuideline.x` property
        getter.

        :return: An :class:`int` or a :class:`flat` representing the
            x-coordinate of the guideline. The value will be normalized
            with :func:`normalizers.normalizeX`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_x(self, value: IntFloatType) -> None:
        """Set the native guideline's x-coordinate.

        This is the environment implementation of the :attr:`BaseGuideline.x` property
        setter.

        :param value: The x-coordinate to set as an :class:`int` or a :class:`float`.
            The value will have been normalized with :func:`normalizers.normalizeX`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # y

    y: dynamicProperty = dynamicProperty(
        "base_y",
        """Get or set the guideline's y-coordinate.

        The value must be an :class:`int` or a :class:`flat`.

        :return: An :class:`int` or a :class:`flat` representing the
            y-coordinate of the guideline.

        Example::

            >>> guideline.y
            100
            >>> guideline.y = 101

        """,
    )

    def _get_base_y(self) -> IntFloatType:
        value = self._get_y()
        if value is None:
            return 0
        value = normalizers.normalizeY(value)
        return value

    def _set_base_y(self, value: IntFloatType) -> None:
        if value is None:
            value = 0
        else:
            value = normalizers.normalizeY(value)
        self._set_y(value)

    def _get_y(self) -> IntFloatType:
        """Get the native guideline's y-coordinate.

        This is the environment implementation of the :attr:`BaseGuideline.y` property
        getter.

        :return: An :class:`int` or a :class:`flat` representing the
            y-coordinate of the guideline. The value will be normalized
            with :func:`normalizers.normalizeY`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_y(self, value: IntFloatType) -> None:
        """Set the native guideline's y-coordinate.

        This is the environment implementation of the :attr:`BaseGuideline.y` property
        setter.

        :param value: The y-coordinate to set as an :class:`int` or a :class:`float`.
            The value will have been normalized with :func:`normalizers.normalizeY`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # angle

    angle: dynamicProperty = dynamicProperty(
        "base_angle",
        """Get or set the guideline's angle.

        The value must be :class:`int`, :class:`float` or :obj:`None`.
        If set to :obj:`None`, the angle is automatically derived based on
        the guideline's :attr:`x` and :attr:`y` values:

        - If both :attr:`x` and :attr:`y` are 0, the angle defaults to ``0.0``.
        - If :attr:`x` is 0 and :attr:`y` is not, the angle is ``90.0``.
        - If :attr:`y` is 0 and :attr:`x` is not, the angle is ``0.0``.

        :return: A :class:`float` representing the angle of the guideline.

        Example::

            >>> guideline.angle
            45.0
            >>> guideline.angle = 90

        """,
    )

    def _get_base_angle(self) -> float:
        value = self._get_angle()
        if value is None:
            if self._get_x() != 0 and self._get_y() != 0:
                value = 0
            elif self._get_x() != 0 and self._get_y() == 0:
                value = 90
            elif self._get_x() == 0 and self._get_y() != 0:
                value = 0
            else:
                value = 0
        value = normalizers.normalizeRotationAngle(value)
        return value

    def _set_base_angle(self, value: IntFloatType | None) -> None:
        if value is None:
            if self._get_x() != 0 and self._get_y() != 0:
                value = 0
            elif self._get_x() != 0 and self._get_y() == 0:
                value = 90
            elif self._get_x() == 0 and self._get_y() != 0:
                value = 0
            else:
                value = 0
        value = normalizers.normalizeRotationAngle(value)
        self._set_angle(value)

    def _get_angle(self) -> IntFloatType | None:
        """Get the native guideline's angle.

        This is the environment implementation of the :attr:`BaseGuideline.angle`
        property getter.

        :return: An :class:`int` or a :class:`float` representing the angle of
            the guideline, or :obj:`None`. The value will be normalized
            with :func:`normalizers.normalizeRotationAngle`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_angle(self, value: float | None) -> None:
        """Set the native guideline's angle.

        This is the environment implementation of the :attr:`BaseGuideline.angle`
        property setter.

        :param value: The angle to set as an :class:`int` or a :class:`float`,
            or :obj:`None`. The value will have been normalized
            with :func:`normalizers.normalizeRotationAngle`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # --------------
    # Identification
    # --------------

    # index

    index: dynamicProperty = dynamicProperty(
        "base_index",
        """Get the guideline's index.

        This property is read-only.

        :return: An :class:`int` representing the index of the guideline within
            the ordered list of the parent glyph's guidelines.

        Example::

            >>> guideline.index
            0

        """,
    )

    def _get_base_index(self) -> int | None:
        value = self._get_index()
        value = normalizers.normalizeIndex(value)
        return value

    def _get_index(self) -> int | None:
        """Get the native guideline's index.

        This is the environment implementation of the :attr:`BaseGuideline.index`
        property getter.

        :return: An :class:`int` representing the index of the guideline within
            the ordered list of the parent glyph's guidelines. The value will be
            normalized with :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        glyph = self.glyph
        if glyph is not None:
            parent = glyph
        else:
            parent = self.font
        if parent is None:
            return None
        return parent.guidelines.index(self)

    # name

    name: dynamicProperty = dynamicProperty(
        "base_name",
        """Get or set the guideline's name.

        The value must be a :class:`str` or :obj:`None`.

        :return: A :class:`str` representing the name of the guideline, or :obj:`None`.

            >>> guideline.name
            'my guideline'
            >>> guideline.name = None

        """,
    )

    def _get_base_name(self) -> str | None:
        value = self._get_name()
        if value is not None:
            value = normalizers.normalizeGuidelineName(value)
        return value

    def _set_base_name(self, value: str | None) -> None:
        if value is not None:
            value = normalizers.normalizeGuidelineName(value)
        self._set_name(value)

    def _get_name(self) -> str | None:
        """Get the native guideline's name.

        This is the environment implementation of the :attr:`BaseGuideline.name`
        property getter.

        :return: A :class:`str` representing the name of the guideline,
            or :obj:`None`. The value will have been normalized
            with :func:`normalizers.normalizeGuidelineName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_name(self, value: str | None) -> None:
        """Set the native guideline's name.

        This is the environment implementation of the :attr:`BaseGuideline.name`
        property setter.

        :param value: The name to set as a :class:`str` or :obj:`None`. The
            value will have been normalized
            with :func:`normalizers.normalizeGuidelineName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # color

    color: dynamicProperty = dynamicProperty(
        "base_color",
        """Get or set the guideline's color.

        The value must be a :ref:`type-color` or :obj:`None`.

        :return: A :class:`Color` instance representing the color of the guideline,
            or :obj:`None`.

        Example::

            >>> guideline.color
            None
            >>> guideline.color = (1, 0, 0, 0.5)

        """,
    )

    def _get_base_color(self) -> RGBA | None:
        value = self._get_color()
        if value is not None:
            value = Color(value)
        return value

    def _set_base_color(self, value: RGBALike | None) -> None:
        if value is not None:
            value = normalizers.normalizeColor(value)
        self._set_color(value)

    def _get_color(self) -> RGBA | None:
        """Get the native guideline's color.

        This is the environment implementation of the :attr:`BaseGuideline.color`
        property getter.

        :return: A :ref:`type-color` representing the color of the guideline,
            or :obj:`None`. The value will be normalized
            with :func:`normalizers.normalizeColor`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_color(self, value: RGBALike | None) -> None:
        """Set the native guideline's color.

        This is the environment implementation of the :attr:`BaseGuideline.color`
        property setter.

        :param value: The :ref:`type-color` to set for the guideline or :obj:`None`.
            The value will have been normalized with :func:`normalizers.normalizeColor`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the native guideline according to the given matrix.

        This is the environment implementation of :meth:`BaseGuideline.transformBy`.

        :param matrix: The :ref:`type-transformation` to apply. The value will
             be normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        t = transform.Transform(*matrix)
        # coordinates
        x, y = t.transformPoint((self.x, self.y))
        self.x = x
        self.y = y
        # angle
        angle = math.radians(-self.angle)
        dx = math.cos(angle)
        dy = math.sin(angle)
        tdx, tdy = t.transformPoint((dx, dy))
        ta = math.atan2(tdy - t[5], tdx - t[4])
        self.angle = -math.degrees(ta)

    # -------------
    # Interpolation
    # -------------

    compatibilityReporterClass = GuidelineCompatibilityReporter

    def isCompatible(
        self, other: BaseGuideline, cls=None
    ) -> tuple[bool, GuidelineCompatibilityReporter]:
        """Evaluate interpolation compatibility with another guideline.

        :param other: The other :class:`BaseGuideline` instance to check
            compatibility with.
        :return: A :class:`tuple` where the first element is a :class:`bool`
            indicating compatibility, and the second element is
            a :class:`fontParts.base.compatibility.GuidelineCompatibilityReporter`
            instance.

        Example::

            >>> compatible, report = self.isCompatible(otherGuideline)
            >>> compatible
            True
            >>> compatible
            [Warning] Guideline: "xheight" + "cap_height"
            [Warning] Guideline: "xheight" has name xheight | "cap_height" has
                                  name cap_height

        """
        return super().isCompatible(other, BaseGuideline)

    def _isCompatible(
        self, other: BaseGuideline, reporter: GuidelineCompatibilityReporter
    ) -> None:
        """Evaluate interpolation compatibility with another native guideline.

        This is the environment implementation of :meth:`BaseGuideline.isCompatible`.

        :param other: The other :class:`BaseGuideline` instance to check
            compatibility with.
        :param reporter: An object used to report compatibility issues.

        .. note::

            Subclasses may override this method.

        """
        guideline1 = self
        guideline2 = other
        # guideline names
        if guideline1.name != guideline2.name:
            reporter.nameDifference = True
            reporter.warning = True

    # -------------
    # Normalization
    # -------------

    def round(self) -> None:
        """Round the guideline's coordinate.

        This applies to:

        - :attr:`x`
        - :attr:`y

        It does not apply to :attr:`angle`.

        Example::`

            >>> guideline.round()


        """
        self._round()

    def _round(self, **kwargs: Any) -> None:
        r"""Round the native guideline's coordinate.

        This is the environment implementation of :meth:`BaseGuideline.round`.

        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        x = self._get_x()
        if x is not None:
            self.x = normalizers.normalizeVisualRounding(normalizers.normalizeX(x))
        y = self._get_y()
        if y is not None:
            self.y = normalizers.normalizeVisualRounding(normalizers.normalizeY(y))



from __future__ import annotations

import defcon
from fontParts.base import BaseFeatures
from fontParts.fontshell.base import RBaseObject


class RFeatures(RBaseObject, BaseFeatures):
    wrapClass = defcon.Features

    def _get_text(self) -> str | None:
        features = self.naked()
        return features.text

    def _set_text(self, value: str | None) -> None:
        features = self.naked()
        features.text = value



import unittest
import collections
from fontParts.base import FontPartsError
from .test_image import testImageData


class TestGlyph(unittest.TestCase):
    def getGlyph_generic(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph.name = "Test Glyph 1"
        glyph.unicode = int(ord("X"))
        glyph.width = 250
        glyph.height = 750
        pen = glyph.getPen()
        pen.moveTo((100, -10))
        pen.lineTo((100, 100))
        pen.lineTo((200, 100))
        pen.lineTo((200, 0))
        pen.closePath()
        pen.moveTo((110, 10))
        pen.lineTo((110, 90))
        pen.lineTo((190, 90))
        pen.lineTo((190, 10))
        pen.closePath()
        glyph.appendAnchor("Test Anchor 1", (1, 2))
        glyph.appendAnchor("Test Anchor 2", (3, 4))
        glyph.appendGuideline((1, 2), 0, "Test Guideline 1")
        glyph.appendGuideline((3, 4), 90, "Test Guideline 2")
        return glyph

    def getGlyph_empty(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph.name = "Test Glyph 2"
        glyph.unicode = int(ord("X"))
        glyph.width = 0
        glyph.height = 0
        return glyph

    def get_generic_object(self, obj_name):
        fp_object, _ = self.objectGenerator(obj_name)
        return fp_object

    # -------
    # Parents
    # -------

    def test_get_layer(self):
        font = self.get_generic_object("font")
        layer = font.layers[0]
        glyph = layer.newGlyph("A")
        self.assertEqual(glyph.layer, layer)

    def test_get_layer_orphan_glyph(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.layer)

    def test_get_font(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("A")
        self.assertEqual(glyph.font, font)

    def test_get_font_orphan_glyph(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.font)

    # --------------
    # Identification
    # --------------

    def test_get_name(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(glyph.name, "Test Glyph 1")

    def test_get_name_not_set(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.name)

    def test_set_name_valid(self):
        glyph = self.getGlyph_generic()
        name = "Test Glyph 1"  # the name is intentionally the same
        glyph.name = name
        self.assertEqual(glyph.name, name)

    def test_set_name_invalid(self):
        invalid_names = (
            ("", ValueError),
            ("A", ValueError),  # a glyph with this name already exists
            (3, TypeError),
            (None, TypeError),
        )
        font = self.get_generic_object("font")
        font.newGlyph("A")
        glyph = font.newGlyph("B")
        for name, err in invalid_names:
            with self.assertRaises(err):
                glyph.name = name

    def test_get_unicode(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(glyph.unicode, 88)

    def test_get_unicode_not_set(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.unicode)

    def test_set_unicode_valid(self):
        valid_uni_values = (100, None, 0x6D, "6D")
        for value in valid_uni_values:
            glyph = self.get_generic_object("glyph")
            glyph.unicode = value
            result = int(value, 16) if isinstance(value, str) else value
            self.assertEqual(glyph.unicode, result)

    def test_set_unicode_value(self):
        glyph = self.get_generic_object("glyph")
        glyph.unicodes = (10, 20)
        glyph.unicode = 20
        self.assertEqual(glyph.unicodes, (20,))

    def test_set_unicode_value_none(self):
        glyph = self.get_generic_object("glyph")
        glyph.unicodes = (10, 20)
        glyph.unicode = None
        self.assertEqual(glyph.unicodes, ())

    def test_set_unicode_invalid(self):
        invalid_uni_values = (("GG", ValueError), (True, TypeError), ([], TypeError))
        glyph = self.get_generic_object("glyph")
        for value, err in invalid_uni_values:
            with self.assertRaises(err):
                glyph.unicode = value

    def test_get_unicodes(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(glyph.unicodes, (88,))

    def test_get_unicodes_not_set(self):
        glyph = self.get_generic_object("glyph")
        self.assertEqual(glyph.unicodes, ())

    def test_set_unicodes_valid(self):
        valid_uni_values = ([100, 200], [], (300,), ())
        for values in valid_uni_values:
            glyph = self.get_generic_object("glyph")
            glyph.unicodes = values
            self.assertEqual(glyph.unicodes, tuple(values))

    def test_set_unicodes_invalid(self):
        invalid_uni_values = (("GG", ValueError), (True, TypeError), (30, TypeError))
        glyph = self.get_generic_object("glyph")
        for value, err in invalid_uni_values:
            with self.assertRaises(err):
                glyph.unicodes = value

    def test_set_unicodes_duplicates(self):
        glyph = self.get_generic_object("glyph")
        with self.assertRaises(ValueError):
            glyph.unicodes = (200, 110, 110)

    # -------
    # Metrics
    # -------

    # The methods are dynamically generated by test_generator()

    # Methods to test that empty glyphs have margins of None

    def test_get_leftMargin_not_set(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.leftMargin)

    def test_get_rightMargin_not_set(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.rightMargin)

    def test_get_bottomMargin_not_set(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.bottomMargin)

    def test_get_topMargin_not_set(self):
        glyph = self.get_generic_object("glyph")
        self.assertIsNone(glyph.topMargin)

    # -------
    # Queries
    # -------

    def test_get_bounds(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(glyph.bounds, (100, -10, 200, 100))

    # ------
    # Layers
    # ------

    def test_get_layers(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("A")
        layers = glyph.layers
        self.assertEqual(len(layers), 1)
        self.assertEqual(glyph.layer.name, font.defaultLayerName)
        self.assertEqual(
            layers[0],
            glyph,  # a glyph layer is really just a glyph
        )
        self.assertEqual(layers[0].name, "A")

    def test_get_layers_orphan_glyph(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(glyph.layers, ())

    def test_getLayer_valid(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("B")
        self.assertEqual(glyph.getLayer(font.defaultLayerName).name, "B")

    def test_getLayer_valid_not_found(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("B")
        with self.assertRaises(ValueError):
            # No layer named 'layer_name' in glyph 'B'
            glyph.getLayer("layer_name")

    def test_getLayer_invalid(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("B")
        with self.assertRaises(TypeError):
            glyph.getLayer()
        with self.assertRaises(TypeError):
            glyph.getLayer(None)
        with self.assertRaises(TypeError):
            glyph.getLayer(0)
        with self.assertRaises(ValueError):
            # Layer names must be at least one character long
            glyph.getLayer("")

    def test_newLayer_valid(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("C")
        self.assertEqual(len(glyph.layers), 1)
        layer = glyph.newLayer("background")
        self.assertEqual(len(glyph.layers), 2)
        self.assertEqual(layer.name, "C")

    def test_newLayer_valid_already_exists(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("C")
        self.assertEqual(len(glyph.layers), 1)
        glyph.newLayer("mask")
        self.assertEqual(len(glyph.layers), 2)
        glyph.newLayer("mask")  # intentional duplicate line
        self.assertEqual(len(glyph.layers), 2)

    def test_newLayer_invalid(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("C")
        with self.assertRaises(TypeError):
            glyph.newLayer()
        with self.assertRaises(TypeError):
            glyph.newLayer(0)
        with self.assertRaises(ValueError):
            # Layer names must be at least one character long
            glyph.newLayer("")

    def test_removeLayer_valid_type_string(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("D")
        self.assertEqual(len(glyph.layers), 1)
        glyph.removeLayer(font.defaultLayerName)
        self.assertEqual(len(glyph.layers), 0)

    def test_removeLayer_valid_type_glyph_layer(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("D")
        self.assertEqual(len(glyph.layers), 1)
        glyph.removeLayer(glyph.layers[0])
        self.assertEqual(len(glyph.layers), 0)

    def test_removeLayer_valid_not_found_type_string(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("D")
        with self.assertRaises(ValueError):
            # No layer named 'layer_name' in glyph 'D'
            glyph.removeLayer("layer_name")

    def test_removeLayer_invalid(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("D")
        with self.assertRaises(TypeError):
            glyph.removeLayer()
        with self.assertRaises(TypeError):
            glyph.removeLayer(0)
        with self.assertRaises(ValueError):
            # Layer names must be at least one character long
            glyph.removeLayer("")

    # ------
    # Global
    # ------

    def test_clear(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        self.assertEqual(len(glyph), 2)
        self.assertEqual(len(glyph.components), 1)
        self.assertEqual(len(glyph.anchors), 2)
        self.assertEqual(len(glyph.guidelines), 2)
        glyph.clear(
            contours=False,
            components=False,
            anchors=False,
            guidelines=False,
            image=False,
        )
        glyph.clear()
        self.assertEqual(len(glyph), 0)
        self.assertEqual(len(glyph.components), 0)
        self.assertEqual(len(glyph.anchors), 0)
        self.assertEqual(len(glyph.guidelines), 0)

    def test_appendGlyph(self):
        glyph_one = self.getGlyph_generic()
        glyph_two = self.getGlyph_generic()
        glyph_one.appendComponent("component 1")
        glyph_two.appendComponent("component 2")
        self.assertEqual(len(glyph_one), 2)
        self.assertEqual(len(glyph_one.components), 1)
        self.assertEqual(len(glyph_one.anchors), 2)
        self.assertEqual(len(glyph_one.guidelines), 2)
        glyph_one.appendGlyph(glyph_two)
        glyph_one.appendGlyph(glyph_two, (300, -40))
        self.assertEqual(len(glyph_one), 6)
        self.assertEqual(len(glyph_one.components), 3)
        self.assertEqual(len(glyph_one.anchors), 6)
        self.assertEqual(len(glyph_one.guidelines), 6)

    # --------
    # Contours
    # --------

    def test_get_contour_invalid(self):
        glyph = self.getGlyph_generic()
        with self.assertRaises(ValueError):
            # No contour located at index 5
            glyph[5]

    def test_insertContour_offset_valid(self):
        glyph = self.getGlyph_generic()
        contour = self.get_generic_object("contour")
        index = 1
        contour.insertPoint(0, position=(0, 0))
        contour.insertPoint(1, position=(100, 100))
        contour.insertPoint(2, position=(0, 100))

        self.assertEqual(len(glyph), 2)
        newcontour = glyph.insertContour(index, contour, (45, 50))
        self.assertEqual(len(glyph), 3)

        self.assertEqual(newcontour, glyph[index])
        self.assertEqual(len(newcontour.points), 3)
        self.assertEqual(newcontour.points[0].x, 45)
        self.assertEqual(newcontour.points[0].y, 50)

    def test_appendContour_offset_valid(self):
        glyph = self.getGlyph_generic()
        contour = self.get_generic_object("contour")
        contour.insertPoint(0, position=(0, 0))
        contour.insertPoint(1, position=(100, 100))
        contour.insertPoint(2, position=(0, 100))

        self.assertEqual(len(glyph), 2)
        newcontour = glyph.appendContour(contour, (45, 50))
        self.assertEqual(len(glyph), 3)

        self.assertEqual(newcontour, glyph[-1])
        self.assertEqual(len(newcontour.points), 3)
        self.assertEqual(newcontour.points[0].x, 45)
        self.assertEqual(newcontour.points[0].y, 50)

    def test_removeContour_valid(self):
        glyph = self.getGlyph_generic()
        contour = self.get_generic_object("contour")
        glyph.appendContour(contour)
        contour1 = glyph.contours[1]
        self.assertEqual(len(glyph), 3)
        glyph.removeContour(contour1)
        self.assertEqual(len(glyph), 2)
        glyph.removeContour(0)
        self.assertEqual(len(glyph), 1)

    def test_removeContour_invalid(self):
        glyph = self.getGlyph_generic()
        with self.assertRaises(ValueError):
            # No contour located at index 5
            glyph.removeContour(5)
        with self.assertRaises(FontPartsError):
            # The contour could not be found
            glyph.removeContour(self.get_generic_object("contour"))

    def test_clearContours(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(len(glyph), 2)
        glyph.clearContours()
        self.assertEqual(len(glyph), 0)

    def test_autoContourOrder_points(self):
        glyph = self.getGlyph_empty()
        pen = glyph.getPen()
        pen.moveTo((287, 212))
        pen.lineTo((217, 108))
        pen.lineTo((407, 109))
        pen.closePath()

        pen = glyph.getPen()
        pen.moveTo((73, 184))
        pen.lineTo((39, 112))
        pen.lineTo((147, 61))
        pen.lineTo((140, 137))
        pen.lineTo((110, 176))
        pen.closePath()

        pen = glyph.getPen()
        pen.moveTo((60, 351))
        pen.lineTo((149, 421))
        pen.lineTo((225, 398))
        pen.lineTo((237, 290))
        pen.lineTo((183, 239))
        pen.lineTo((129, 245))
        pen.lineTo((70, 285))
        pen.closePath()

        glyph.autoContourOrder()
        self.assertEqual([len(c.points) for c in glyph.contours], [7, 5, 3])

    def test_autoContourOrder_segments(self):
        glyph = self.getGlyph_empty()

        pen = glyph.getPen()
        pen.moveTo((116, 202))
        pen.curveTo((116, 308), (156, 348), (245, 348))
        pen.closePath()

        pen = glyph.getPen()
        pen.moveTo((261, 212))
        pen.lineTo((335, 212))
        pen.lineTo((335, 301))
        pen.lineTo((261, 301))
        pen.closePath()

        glyph.autoContourOrder()
        self.assertEqual([len(c.segments) for c in glyph.contours], [4, 2])

    def test_autoContourOrder_fuzzycenter(self):
        glyph = self.getGlyph_empty()

        # the contours are overlapping too much
        # the different position of their center points
        # should not matter
        pen = glyph.getPen()
        pen.moveTo((313, 44))
        pen.curveTo((471, 44), (600, 174), (600, 332))
        pen.curveTo((600, 490), (471, 619), (313, 619))
        pen.curveTo((155, 619), (26, 490), (26, 332))
        pen.curveTo((26, 174), (155, 44), (313, 44))
        pen.closePath()

        pen = glyph.getPen()
        pen.moveTo((288, 122))
        pen.curveTo((383, 122), (461, 200), (461, 295))
        pen.curveTo((461, 390), (383, 468), (288, 468))
        pen.curveTo((192, 468), (114, 390), (114, 295))
        pen.curveTo((114, 200), (192, 122), (288, 122))
        pen.closePath()

        glyph.autoContourOrder()

        self.assertTrue(
            glyph.contours[0].points[0].x == 313
            and glyph.contours[1].points[0].x == 288
        )

    def test_autoContourOrder_distantCenter(self):
        glyph = self.getGlyph_empty()

        # both outlines have same structure
        # but their center points are far away
        # so, the center points should inform
        # the sorting
        # from left to right first, then bottom to top
        pen = glyph.getPen()
        pen.moveTo((313, 44))
        pen.curveTo((471, 44), (600, 174), (600, 332))
        pen.curveTo((600, 490), (471, 619), (313, 619))
        pen.curveTo((155, 619), (26, 490), (26, 332))
        pen.curveTo((26, 174), (155, 44), (313, 44))
        pen.closePath()

        pen = glyph.getPen()
        pen.moveTo((-142, -118))
        pen.curveTo((-47, -118), (31, -40), (31, 55))
        pen.curveTo((31, 150), (-47, 228), (-142, 228))
        pen.curveTo((-238, 228), (-316, 150), (-316, 55))
        pen.curveTo((-316, -40), (-238, -118), (-142, -118))
        pen.closePath()

        glyph.autoContourOrder()
        self.assertTrue(
            glyph.contours[0].points[0].x == -142
            and glyph.contours[1].points[0].x == 313
        )

    def test_autoContourOrder_bboxsurface(self):
        glyph = self.getGlyph_empty()

        # both outlines share the same center
        # so, the surface of the bounding box should inform
        # the sorting, from larger to smaller
        pen = glyph.getPen()
        pen.moveTo((100, 50))
        pen.curveTo((128, 50), (150, 72), (150, 100))
        pen.curveTo((150, 128), (128, 150), (100, 150))
        pen.curveTo((72, 150), (50, 128), (50, 100))
        pen.curveTo((50, 72), (72, 50), (100, 50))
        pen.closePath()

        pen = glyph.getPen()
        pen.moveTo((100, 0))
        pen.curveTo((155, 0), (200, 45), (200, 100))
        pen.curveTo((200, 155), (155, 200), (100, 200))
        pen.curveTo((45, 200), (0, 155), (0, 100))
        pen.curveTo((0, 45), (45, 0), (100, 0))
        pen.closePath()

        glyph.autoContourOrder()
        self.assertTrue(
            glyph.contours[0].points[0].y == 0 and glyph.contours[1].points[0].y == 50
        )

    # ----------
    # Components
    # ----------

    # appendComponent

    def test_appendComponent_invalid_circularReference(self):
        glyph = self.getGlyph_generic()
        with self.assertRaises(FontPartsError):
            glyph.appendComponent(glyph.name)

    def test_appendComponent_valid_object(self):
        glyph = self.getGlyph_generic()
        src, _ = self.objectGenerator("component")
        src.baseGlyph = "test"
        src.transformation = (1, 2, 3, 4, 5, 6)
        src.getIdentifier()
        dst = glyph.appendComponent(component=src)
        self.assertNotEqual(src, dst)
        self.assertEqual(src.baseGlyph, dst.baseGlyph)
        self.assertEqual(src.transformation, dst.transformation)
        self.assertEqual(src.identifier, dst.identifier)

    def test_appendComponent_valid_object_baseGlyph(self):
        glyph = self.getGlyph_generic()
        src, _ = self.objectGenerator("component")
        src.baseGlyph = "assigned"
        dst = glyph.appendComponent(component=src, baseGlyph="argument")
        self.assertEqual(dst.baseGlyph, "argument")

    def test_appendComponent_valid_object_offset(self):
        glyph = self.getGlyph_generic()
        src, _ = self.objectGenerator("component")
        src.baseGlyph = "test"
        src.transformation = (1, 2, 3, 4, 5, 6)
        dst = glyph.appendComponent(component=src, offset=(-1, -2))
        self.assertEqual(dst.offset, (-1, -2))
        self.assertEqual(dst.transformation, (1, 2, 3, 4, -1, -2))

    def test_appendComponent_valid_object_scale(self):
        glyph = self.getGlyph_generic()
        src, _ = self.objectGenerator("component")
        src.baseGlyph = "test"
        src.transformation = (1, 2, 3, 4, 5, 6)
        dst = glyph.appendComponent(component=src, scale=(-1, -2))
        self.assertEqual(dst.scale, (-1, -2))
        self.assertEqual(dst.transformation, (-1, 2, 3, -2, 5, 6))

    def test_appendComponent_valid_object_conflictingIdentifier(self):
        glyph = self.getGlyph_generic()
        c = glyph.appendComponent("test")
        existingIdentifier = c.getIdentifier()
        src, _ = self.objectGenerator("component")
        src.baseGlyph = "test"
        src._setIdentifier(existingIdentifier)
        dst = glyph.appendComponent(component=src)
        self.assertNotEqual(src.identifier, dst.identifier)

    def test_appendComponent_invalid_object_circularReference(self):
        glyph = self.getGlyph_generic()
        src, _ = self.objectGenerator("component")
        src.baseGlyph = glyph.name
        with self.assertRaises(FontPartsError):
            glyph.appendComponent(glyph.name)

    # removeComponent

    def test_removeComponent_valid(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        glyph.appendComponent("component 2")
        glyph.appendComponent("component 3")
        self.assertEqual(len(glyph.components), 3)
        component = glyph.components[1]
        glyph.removeComponent(component)
        self.assertEqual(len(glyph.components), 2)
        glyph.removeComponent(0)
        self.assertEqual(len(glyph.components), 1)

    def test_removeComponent_invalid(self):
        glyph = self.getGlyph_generic()
        with self.assertRaises(ValueError):
            # No component located at index 8
            glyph.removeComponent(8)
        with self.assertRaises(FontPartsError):
            # The component could not be found
            glyph.removeComponent(self.get_generic_object("component"))

    # clearComponents

    def test_clearComponents(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        self.assertEqual(len(glyph.components), 1)
        glyph.clearComponents()
        self.assertEqual(len(glyph.components), 0)

    # -------
    # Anchors
    # -------

    # appendAnchor

    def test_appendAnchor_valid_object(self):
        glyph = self.getGlyph_generic()
        src, _ = self.objectGenerator("anchor")
        src.name = "test"
        src.position = (1, 2)
        src.color = (1, 1, 1, 1)
        src.getIdentifier()
        dst = glyph.appendAnchor(anchor=src)
        self.assertNotEqual(src, dst)
        self.assertEqual(src.name, dst.name)
        self.assertEqual(src.position, dst.position)
        self.assertEqual(src.color, dst.color)
        self.assertEqual(src.identifier, dst.identifier)

    # removeAnchor

    def test_removeAnchor_valid(self):
        glyph = self.getGlyph_generic()
        glyph.appendAnchor("base", (250, 0), (1, 0, 0, 0.5))
        anchor = glyph.anchors[1]
        self.assertEqual(len(glyph.anchors), 3)
        glyph.removeAnchor(anchor)
        self.assertEqual(len(glyph.anchors), 2)
        glyph.removeAnchor(0)
        self.assertEqual(len(glyph.anchors), 1)

    def test_removeAnchor_invalid(self):
        glyph = self.getGlyph_generic()
        with self.assertRaises(ValueError):
            # No anchor located at index 4
            glyph.removeAnchor(4)
        with self.assertRaises(FontPartsError):
            # The anchor could not be found
            glyph.removeAnchor(self.get_generic_object("anchor"))

    # clearAnchors

    def test_clearAnchors(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(len(glyph.anchors), 2)
        glyph.clearAnchors()
        self.assertEqual(len(glyph.anchors), 0)

    # ----------
    # Guidelines
    # ----------

    # appendGuideline

    def test_appendGuideline_valid_object(self):
        glyph = self.getGlyph_generic()
        src, _ = self.objectGenerator("guideline")
        src.position = (1, 2)
        src.angle = 123
        src.name = "test"
        src.color = (1, 1, 1, 1)
        src.getIdentifier()
        dst = glyph.appendGuideline(guideline=src)
        self.assertNotEqual(src, dst)
        self.assertEqual(src.position, dst.position)
        self.assertEqual(src.angle, dst.angle)
        self.assertEqual(src.name, dst.name)
        self.assertEqual(src.color, dst.color)
        self.assertEqual(src.identifier, dst.identifier)

    # removeGuideline

    def test_removeGuideline_valid(self):
        glyph = self.getGlyph_generic()
        glyph.appendGuideline((5, -10), 90, None, (1, 0, 0, 0.5))
        guideline = glyph.guidelines[1]
        self.assertEqual(len(glyph.guidelines), 3)
        glyph.removeGuideline(guideline)
        self.assertEqual(len(glyph.guidelines), 2)
        glyph.removeGuideline(0)
        self.assertEqual(len(glyph.guidelines), 1)

    def test_removeGuideline_invalid(self):
        glyph = self.getGlyph_generic()
        with self.assertRaises(ValueError):
            # No guideline located at index 6
            glyph.removeGuideline(6)
        with self.assertRaises(FontPartsError):
            # The guideline could not be found
            glyph.removeGuideline(self.get_generic_object("guideline"))

    # clearGuidelines

    def test_clearGuidelines(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(len(glyph.guidelines), 2)
        glyph.clearGuidelines()
        self.assertEqual(len(glyph.guidelines), 0)

    # -----
    # Image
    # -----

    def test_addImage(self):
        font = self.get_generic_object("font")
        glyph = font.newGlyph("glyphWithImage")
        image = glyph.addImage(data=testImageData)
        self.assertEqual(image.data, testImageData)

    # ----
    # Hash
    # ----

    def test_hash_object_self(self):
        glyph_one = self.getGlyph_generic()
        glyph_one.name = "Test"
        self.assertEqual(hash(glyph_one), hash(glyph_one))

    def test_hash_object_other(self):
        glyph_one = self.getGlyph_generic()
        glyph_two = self.getGlyph_generic()
        glyph_one.name = "Test"
        glyph_two.name = "Test"
        self.assertNotEqual(hash(glyph_one), hash(glyph_two))

    def test_hash_object_self_variable_assignment(self):
        glyph_one = self.getGlyph_generic()
        a = glyph_one
        self.assertEqual(hash(glyph_one), hash(a))

    def test_hash_object_other_variable_assignment(self):
        glyph_one = self.getGlyph_generic()
        glyph_two = self.getGlyph_generic()
        a = glyph_one
        self.assertNotEqual(hash(glyph_two), hash(a))

    def test_is_hashable(self):
        glyph_one = self.getGlyph_generic()
        self.assertTrue(isinstance(glyph_one, collections.abc.Hashable))

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        glyph_one = self.getGlyph_generic()
        glyph_one.name = "Test"
        self.assertEqual(glyph_one, glyph_one)

    def test_object_not_equal_other(self):
        glyph_one = self.getGlyph_generic()
        glyph_two = self.getGlyph_generic()
        self.assertNotEqual(glyph_one, glyph_two)

    def test_object_not_equal_other_name_same(self):
        glyph_one = self.getGlyph_generic()
        glyph_two = self.getGlyph_generic()
        glyph_one.name = "Test"
        glyph_two.name = "Test"
        self.assertNotEqual(glyph_one, glyph_two)

    def test_object_equal_variable_assignment(self):
        glyph_one = self.getGlyph_generic()
        a = glyph_one
        a.name = "Other"
        self.assertEqual(glyph_one, a)

    def test_object_not_equal_variable_assignment(self):
        glyph_one = self.getGlyph_generic()
        glyph_two = self.getGlyph_generic()
        a = glyph_one
        self.assertNotEqual(glyph_two, a)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        glyph = self.getGlyph_generic()
        try:
            glyph.selected = False
        except NotImplementedError:
            return
        glyph.selected = True
        self.assertTrue(glyph.selected)

    def test_not_selected_false(self):
        glyph = self.getGlyph_generic()
        try:
            glyph.selected = False
        except NotImplementedError:
            return
        self.assertFalse(glyph.selected)

    # Contours

    def test_selectedContours_default(self):
        glyph = self.getGlyph_generic()
        contour1 = glyph.contours[0]
        try:
            contour1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(glyph.selectedContours, ())

    def test_selectedContours_setSubObject(self):
        glyph = self.getGlyph_generic()
        contour1 = glyph.contours[0]
        contour2 = glyph.contours[1]
        try:
            contour1.selected = False
        except NotImplementedError:
            return
        contour2.selected = True
        self.assertEqual(glyph.selectedContours, (contour2,))

    def test_selectedContours_setFilledList(self):
        glyph = self.getGlyph_generic()
        contour1 = glyph.contours[0]
        contour2 = glyph.contours[1]
        try:
            contour1.selected = False
        except NotImplementedError:
            return
        glyph.selectedContours = [contour1, contour2]
        self.assertEqual(glyph.selectedContours, (contour1, contour2))

    def test_selectedContours_setEmptyList(self):
        glyph = self.getGlyph_generic()
        contour1 = glyph.contours[0]
        try:
            contour1.selected = True
        except NotImplementedError:
            return
        glyph.selectedContours = []
        self.assertEqual(glyph.selectedContours, ())

    # Components
    def test_selectedComponents_default(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        component1 = glyph.components[0]
        try:
            component1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(glyph.selectedComponents, ())

    def test_selectedComponents_setSubObject(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        glyph.appendComponent("component 2")
        component1 = glyph.components[0]
        component2 = glyph.components[1]
        try:
            component1.selected = False
        except NotImplementedError:
            return
        component2.selected = True
        self.assertEqual(glyph.selectedComponents, (component2,))

    def test_selectedComponents_setFilledList(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        glyph.appendComponent("component 2")
        component1 = glyph.components[0]
        component2 = glyph.components[1]
        try:
            component1.selected = False
        except NotImplementedError:
            return
        glyph.selectedComponents = [component1, component2]
        self.assertEqual(glyph.selectedComponents, (component1, component2))

    def test_selectedComponents_setEmptyList(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        component1 = glyph.components[0]
        try:
            component1.selected = True
        except NotImplementedError:
            return
        glyph.selectedComponents = []
        self.assertEqual(glyph.selectedComponents, ())

    # Anchors

    def test_selectedAnchors_default(self):
        glyph = self.getGlyph_generic()
        anchor1 = glyph.anchors[0]
        try:
            anchor1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(glyph.selectedAnchors, ())

    def test_selectedAnchors_setSubObject(self):
        glyph = self.getGlyph_generic()
        anchor1 = glyph.anchors[0]
        anchor2 = glyph.anchors[1]
        try:
            anchor1.selected = False
        except NotImplementedError:
            return
        anchor2.selected = True
        self.assertEqual(glyph.selectedAnchors, (anchor2,))

    def test_selectedAnchors_setFilledList(self):
        glyph = self.getGlyph_generic()
        anchor1 = glyph.anchors[0]
        anchor2 = glyph.anchors[1]
        try:
            anchor1.selected = False
        except NotImplementedError:
            return
        glyph.selectedAnchors = [anchor1, anchor2]
        self.assertEqual(glyph.selectedAnchors, (anchor1, anchor2))

    def test_selectedAnchors_setEmptyList(self):
        glyph = self.getGlyph_generic()
        anchor1 = glyph.anchors[0]
        try:
            anchor1.selected = True
        except NotImplementedError:
            return
        glyph.selectedAnchors = []
        self.assertEqual(glyph.selectedAnchors, ())

    # Guidelines

    def test_selectedGuidelines_default(self):
        glyph = self.getGlyph_generic()
        guideline1 = glyph.guidelines[0]
        try:
            guideline1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(glyph.selectedGuidelines, ())

    def test_selectedGuidelines_setSubObject(self):
        glyph = self.getGlyph_generic()
        guideline1 = glyph.guidelines[0]
        guideline2 = glyph.guidelines[1]
        try:
            guideline1.selected = False
        except NotImplementedError:
            return
        guideline2.selected = True
        self.assertEqual(glyph.selectedGuidelines, (guideline2,))

    def test_selectedGuidelines_setFilledList(self):
        glyph = self.getGlyph_generic()
        guideline1 = glyph.guidelines[0]
        guideline2 = glyph.guidelines[1]
        try:
            guideline1.selected = False
        except NotImplementedError:
            return
        glyph.selectedGuidelines = [guideline1, guideline2]
        self.assertEqual(glyph.selectedGuidelines, (guideline1, guideline2))

    def test_selectedGuidelines_setEmptyList(self):
        glyph = self.getGlyph_generic()
        guideline1 = glyph.guidelines[0]
        try:
            guideline1.selected = True
        except NotImplementedError:
            return
        glyph.selectedGuidelines = []
        self.assertEqual(glyph.selectedGuidelines, ())

    # -------------
    # Compatibility
    # -------------

    def test_isCompatible_anchors(self):
        glyph1 = self.getGlyph_generic()
        glyph1.clearAnchors()
        glyph1.appendAnchor("a", (0, 0))
        glyph1.appendAnchor("b", (0, 0))
        glyph2 = self.getGlyph_generic()
        glyph2.clearAnchors()
        glyph2.appendAnchor("a", (0, 0))
        glyph2.appendAnchor("b", (0, 0))
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertTrue(is_compatible)
        self.assertFalse(report.anchorDifferences)
        self.assertFalse(report.anchorOrderDifference)
        self.assertFalse(report.anchorCountDifference)
        self.assertFalse(report.anchorsMissingFromGlyph1)
        self.assertFalse(report.anchorsMissingFromGlyph2)

    def test_isCompatible_anchors_order(self):
        glyph1 = self.getGlyph_generic()
        glyph1.clearAnchors()
        glyph1.appendAnchor("a", (0, 0))
        glyph1.appendAnchor("b", (0, 0))
        glyph2 = self.getGlyph_generic()
        glyph2.clearAnchors()
        glyph2.appendAnchor("b", (0, 0))
        glyph2.appendAnchor("a", (0, 0))
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertTrue(is_compatible)
        self.assertEqual(report.anchorDifferences, [(0, "a", "b"), (1, "b", "a")])
        self.assertTrue(report.anchorOrderDifference)
        self.assertFalse(report.anchorCountDifference)
        self.assertFalse(report.anchorsMissingFromGlyph1)
        self.assertFalse(report.anchorsMissingFromGlyph2)

    def test_isCompatible_anchors_intersecting(self):
        glyph1 = self.getGlyph_generic()
        glyph1.clearAnchors()
        glyph1.appendAnchor("a", (0, 0))
        glyph1.appendAnchor("b", (0, 0))
        glyph2 = self.getGlyph_generic()
        glyph2.clearAnchors()
        glyph2.appendAnchor("a", (0, 0))
        glyph2.appendAnchor("b", (0, 0))
        glyph2.appendAnchor("b", (0, 0))
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertTrue(is_compatible)
        self.assertEqual(report.anchorDifferences, [(2, None, "b")])
        self.assertFalse(report.anchorOrderDifference)
        self.assertTrue(report.anchorCountDifference)
        self.assertEqual(report.anchorsMissingFromGlyph1, ["b"])
        self.assertFalse(report.anchorsMissingFromGlyph2)

    def test_isCompatible_anchors_disjoint(self):
        glyph1 = self.getGlyph_generic()
        glyph1.clearAnchors()
        glyph1.appendAnchor("x", (0, 0))
        glyph2 = self.getGlyph_generic()
        glyph2.clearAnchors()
        glyph2.appendAnchor("a", (0, 0))
        glyph2.appendAnchor("a", (0, 0))
        glyph2.appendAnchor("b", (0, 0))
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertTrue(is_compatible)
        self.assertEqual(
            report.anchorDifferences, [(0, "x", "a"), (1, None, "a"), (2, None, "b")]
        )
        self.assertFalse(report.anchorOrderDifference)
        self.assertTrue(report.anchorCountDifference)
        self.assertEqual(report.anchorsMissingFromGlyph1, ["a", "a", "b"])
        self.assertEqual(report.anchorsMissingFromGlyph2, ["x"])

    def test_isCompatible_components(self):
        glyph1, _ = self.objectGenerator("glyph")
        glyph1.appendComponent("a")
        glyph1.appendComponent("b")
        glyph2, _ = self.objectGenerator("glyph")
        glyph2.appendComponent("a")
        glyph2.appendComponent("b")
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertTrue(is_compatible)
        self.assertFalse(report.componentDifferences)
        self.assertFalse(report.componentOrderDifference)
        self.assertFalse(report.componentCountDifference)
        self.assertFalse(report.componentsMissingFromGlyph1)
        self.assertFalse(report.componentsMissingFromGlyph2)

    def test_isCompatible_components_order(self):
        glyph1, _ = self.objectGenerator("glyph")
        glyph1.appendComponent("a")
        glyph1.appendComponent("b")
        glyph2, _ = self.objectGenerator("glyph")
        glyph2.appendComponent("b")
        glyph2.appendComponent("a")
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertTrue(is_compatible)
        self.assertEqual(report.componentDifferences, [(0, "a", "b"), (1, "b", "a")])
        self.assertTrue(report.componentOrderDifference)
        self.assertFalse(report.componentCountDifference)
        self.assertFalse(report.componentsMissingFromGlyph1)
        self.assertFalse(report.componentsMissingFromGlyph2)

    def test_isCompatible_components_intersecting(self):
        glyph1, _ = self.objectGenerator("glyph")
        glyph1.appendComponent("a")
        glyph1.appendComponent("b")
        glyph2, _ = self.objectGenerator("glyph")
        glyph2.appendComponent("a")
        glyph2.appendComponent("b")
        glyph2.appendComponent("b")
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertFalse(is_compatible)
        self.assertEqual(report.componentDifferences, [(2, None, "b")])
        self.assertFalse(report.componentOrderDifference)
        self.assertTrue(report.componentCountDifference)
        self.assertEqual(report.componentsMissingFromGlyph1, ["b"])
        self.assertFalse(report.componentsMissingFromGlyph2)

    def test_isCompatible_components_disjoint(self):
        glyph1, _ = self.objectGenerator("glyph")
        glyph1.appendComponent("x")
        glyph2, _ = self.objectGenerator("glyph")
        glyph2.appendComponent("a")
        glyph2.appendComponent("a")
        glyph2.appendComponent("b")
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertFalse(is_compatible)
        self.assertEqual(
            report.componentDifferences, [(0, "x", "a"), (1, None, "a"), (2, None, "b")]
        )
        self.assertFalse(report.componentOrderDifference)
        self.assertTrue(report.componentCountDifference)
        self.assertEqual(report.componentsMissingFromGlyph1, ["a", "a", "b"])
        self.assertEqual(report.componentsMissingFromGlyph2, ["x"])

    def test_isCompatible_components_disjoint_equal_size(self):
        glyph1, _ = self.objectGenerator("glyph")
        glyph1.appendComponent("x")
        glyph1.appendComponent("y")
        glyph2, _ = self.objectGenerator("glyph")
        glyph2.appendComponent("a")
        glyph2.appendComponent("b")
        is_compatible, report = glyph1.isCompatible(glyph2)
        self.assertFalse(is_compatible)
        self.assertEqual(report.componentDifferences, [(0, "x", "a"), (1, "y", "b")])
        self.assertFalse(report.componentOrderDifference)
        self.assertFalse(report.componentCountDifference)
        self.assertEqual(report.componentsMissingFromGlyph1, ["a", "b"])
        self.assertEqual(report.componentsMissingFromGlyph2, ["x", "y"])

    # -------------
    # Interpolation
    # -------------

    def test_interpolate_glyphWidth_without_rounding(self):
        interpolated, _ = self.objectGenerator("glyph")
        glyph_min, _ = self.objectGenerator("glyph")
        glyph_max, _ = self.objectGenerator("glyph")
        glyph_min.width = 1000
        glyph_max.width = 2000
        interpolated.interpolate(0.5154, glyph_min, glyph_max, round=False)
        self.assertEqual(interpolated.width, 1515.4)

    def test_interpolate_glyphWidth_with_rounding(self):
        interpolated, _ = self.objectGenerator("glyph")
        glyph_min, _ = self.objectGenerator("glyph")
        glyph_max, _ = self.objectGenerator("glyph")
        glyph_min.width = 1000
        glyph_max.width = 2000
        interpolated.interpolate(0.5154, glyph_min, glyph_max, round=True)
        self.assertEqual(interpolated.width, 1515)

    # ---------------
    # Transformations
    # ---------------

    def test_moveBy_only_contours(self):
        glyph = self.getGlyph_generic()
        glyph.moveBy((100, 0))
        self.assertEqual(glyph.bounds[0], 200)
        glyph.moveBy((0, 250))
        self.assertEqual(glyph.bounds[1], 240)

    # ---
    # API
    # ---

    def test_isEmpty_false_outlines(self):
        glyph = self.getGlyph_generic()
        self.assertFalse(glyph.isEmpty())

    def test_isEmpty_true_clear(self):
        glyph = self.getGlyph_generic()
        glyph.clear()
        self.assertTrue(glyph.isEmpty())

    def test_isEmpty_false_component(self):
        glyph = self.getGlyph_generic()
        glyph.clear()
        glyph.appendComponent("component 1")
        self.assertFalse(glyph.isEmpty())

    def test_removeOverlap(self):
        glyph = self.getGlyph_generic()
        self.assertEqual(len(glyph), 2)
        glyph.removeOverlap()
        self.assertEqual(len(glyph), 1)


def test_generator(test_name, metric, value):
    if "_invalid_" in test_name:

        def test(self):
            glyph = self.getGlyph_generic()
            with self.assertRaises(TypeError):
                setattr(glyph, metric, value)
    else:

        def test(self):
            glyph = self.getGlyph_generic()
            if "_set_" in test_name:
                setattr(glyph, metric, value)
            self.assertEqual(getattr(glyph, metric), value)

    return test


t_names = (
    "_get",
    "_set_valid_positive",
    "_set_valid_negative",
    "_set_valid_zero",
    "_set_valid_float",
    "_set_invalid_string",
    "_set_invalid_none",
)
invalid = ("abc", None)
metrics = {
    "width": (250, 300, -485, 0, 101.5) + invalid,
    "height": (750, 800, -10, 0, 801.5) + invalid,
    "leftMargin": (100, 200, -15, 0, 201.5) + invalid,
    "rightMargin": (50, 80, -20, 0, 81.5) + invalid,
    "bottomMargin": (-10, 150, -35, 0, 151.5) + invalid,
    "topMargin": (650, 750, -250, 0, 751.5) + invalid,
}

for i, t_name_suffix in enumerate(t_names):
    for metric_name, values in metrics.items():
        test_name = f"test_{metric_name}{t_name_suffix}"
        test = test_generator(test_name, metric_name, values[i])
        setattr(TestGlyph, test_name, test)



import unittest
from fontParts.base.bounds import Bounds


class TestBounds(unittest.TestCase):
    def test_new_tuple(self):
        bounds = Bounds((1, 2, 3, 4))
        self.assertEqual(bounds.xMin, 1.0)
        self.assertEqual(bounds.yMin, 2.0)
        self.assertEqual(bounds.xMax, 3.0)
        self.assertEqual(bounds.yMax, 4.0)

    def test_new_args(self):
        bounds = Bounds(1, 2, 3, 4)
        self.assertEqual(bounds.xMin, 1.0)
        self.assertEqual(bounds.yMin, 2.0)
        self.assertEqual(bounds.xMax, 3.0)
        self.assertEqual(bounds.yMax, 4.0)

    def test_new_kwargs(self):
        bounds = Bounds(xMin=1, yMin=2, xMax=3, yMax=4)
        self.assertEqual(bounds.xMin, 1.0)
        self.assertEqual(bounds.yMin, 2.0)
        self.assertEqual(bounds.xMax, 3.0)
        self.assertEqual(bounds.yMax, 4.0)

    def test_width(self):
        bounds = Bounds(1, 2, 3, 4)
        self.assertEqual(bounds.width, 2.0)

    def test_height(self):
        bounds = Bounds(1, 2, 3, 4)
        self.assertEqual(bounds.height, 2.0)

    def test_xCenter(self):
        bounds = Bounds(1, 2, 3, 4)
        self.assertEqual(bounds.xCenter, 2.0)

    def test_yCenter(self):
        bounds = Bounds(1, 2, 3, 4)
        self.assertEqual(bounds.yCenter, 3.0)



import unittest
import collections
from fontParts.base import FontPartsError


class TestPoint(unittest.TestCase):
    def getPoint_generic(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((101, 202), "line")
        point = contour.points[1]
        point.smooth = True
        return point

    # ----
    # repr
    # ----

    def test_reprContents(self):
        point = self.getPoint_generic()
        value = point._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_withName(self):
        point = self.getPoint_withName()
        value = point._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_isSmooth(self):
        point = self.getPoint_generic()
        point.smooth = True
        value = point._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_noContour(self):
        point, _ = self.objectGenerator("point")
        value = point._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    # -------
    # Parents
    # -------

    def test_get_parent_font(self):
        font, _ = self.objectGenerator("font")
        layer = font.newLayer("L")
        glyph = layer.newGlyph("X")
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        glyph.appendContour(contour)
        contour = glyph.contours[0]
        point = contour.points[0]
        self.assertIsNotNone(point.font)
        self.assertEqual(point.font, font)

    def test_get_parent_noFont(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        glyph.appendContour(contour)
        contour = glyph.contours[0]
        point = contour.points[0]
        self.assertIsNone(point.font)

    def test_get_parent_layer(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        glyph.appendContour(contour)
        contour = glyph.contours[0]
        point = contour.points[0]
        self.assertIsNotNone(point.layer)
        self.assertEqual(point.layer, layer)

    def test_get_parent_noLayer(self):
        glyph, _ = self.objectGenerator("glyph")
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        glyph.appendContour(contour)
        contour = glyph.contours[0]
        point = contour.points[0]
        self.assertIsNone(point.font)
        self.assertIsNone(point.layer)

    def test_get_parent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        glyph.appendContour(contour)
        contour = glyph.contours[0]
        point = contour.points[0]
        self.assertIsNotNone(point.glyph)
        self.assertEqual(point.glyph, glyph)

    def test_get_parent_noGlyph(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        point = contour.points[0]
        self.assertIsNone(point.glyph)

    def test_get_parent_contour(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        point = contour.points[0]
        self.assertIsNotNone(point.contour)
        self.assertEqual(point.contour, contour)

    def test_get_parent_noContour(self):
        point, _ = self.objectGenerator("point")
        self.assertIsNone(point.contour)

    def test_get_parent_segment(self):
        point, _ = self.objectGenerator("point")
        with self.assertRaises(AttributeError):
            point.segment

    def test_set_parent_contour(self):
        contour, _ = self.objectGenerator("contour")
        point, _ = self.objectGenerator("point")
        point.contour = contour
        self.assertIsNotNone(point.contour)
        self.assertEqual(point.contour, contour)

    def test_set_already_set_parent_contour(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((10, 20))
        point = contour.points[0]
        contourOther, _ = self.objectGenerator("contour")
        with self.assertRaises(AssertionError):
            point.contour = contourOther

    def test_set_parent_contour_none(self):
        point, _ = self.objectGenerator("point")
        point.contour = None
        self.assertIsNone(point.contour)

    def test_get_parent_glyph_noContour(self):
        point, _ = self.objectGenerator("point")
        self.assertIsNone(point.glyph)

    def test_get_parent_layer_noContour(self):
        point, _ = self.objectGenerator("point")
        self.assertIsNone(point.layer)

    def test_get_parent_font_noContour(self):
        point, _ = self.objectGenerator("point")
        self.assertIsNone(point.font)

    # ----------
    # Attributes
    # ----------

    # type

    def test_get_type(self):
        point = self.getPoint_generic()
        self.assertEqual(point.type, "line")

    def test_set_move(self):
        point = self.getPoint_generic()
        point.type = "move"
        self.assertEqual(point.type, "move")

    def test_set_curve(self):
        point = self.getPoint_generic()
        point.type = "curve"
        self.assertEqual(point.type, "curve")

    def test_set_wcurve(self):
        point = self.getPoint_generic()
        point.type = "qcurve"
        self.assertEqual(point.type, "qcurve")

    def test_set_offcurve(self):
        point = self.getPoint_generic()
        point.type = "offcurve"
        self.assertEqual(point.type, "offcurve")

    def test_set_invalid_point_type_string(self):
        point = self.getPoint_generic()
        with self.assertRaises(ValueError):
            point.type = "xxx"

    def test_set_invalid_point_type_int(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.type = 123

    # smooth

    def test_get_smooth(self):
        point = self.getPoint_generic()
        self.assertEqual(point.smooth, True)

    def test_set_smooth_valid(self):
        point = self.getPoint_generic()
        point.smooth = True
        self.assertEqual(point.smooth, True)

    def test_set_smooth_invalid(self):
        point = self.getPoint_generic()
        with self.assertRaises(ValueError):
            point.smooth = "smooth"

    # x

    def test_get_x(self):
        point = self.getPoint_generic()
        self.assertEqual(point.x, 101)

    def test_set_x_valid_int(self):
        point = self.getPoint_generic()
        point.x = 100
        self.assertEqual(point.x, 100)

    def test_set_x_valid_float(self):
        point = self.getPoint_generic()
        point.x = 100.5
        self.assertEqual(point.x, 100.5)

    def test_set_x_invalidType(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.x = "100"

    # y

    def test_get_y(self):
        point = self.getPoint_generic()
        self.assertEqual(point.y, 202)

    def test_set_y_valid_int(self):
        point = self.getPoint_generic()
        point.y = 200
        self.assertEqual(point.y, 200)

    def test_set_y_valid_float(self):
        point = self.getPoint_generic()
        point.y = 200.5
        self.assertEqual(point.y, 200.5)

    def test_set_y_invalidType(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.y = "200"

    # --------------
    # Identification
    # --------------

    # index

    def getPoint_noParentContour(self):
        point, _ = self.objectGenerator("point")
        point.x = 101
        point.y = 202
        return point

    def test_get_index(self):
        point = self.getPoint_generic()
        self.assertEqual(point.index, 1)

    def test_get_index_noParentContour(self):
        point = self.getPoint_noParentContour()
        self.assertEqual(point.index, None)

    def test_set_index(self):
        point = self.getPoint_generic()
        with self.assertRaises(FontPartsError):
            point.index = 0

    # name

    def getPoint_withName(self):
        point = self.getPoint_generic()
        point.name = "P"
        return point

    def test_get_name_noName(self):
        point = self.getPoint_generic()
        self.assertEqual(point.name, None)

    def test_get_name_hasName(self):
        point = self.getPoint_withName()
        self.assertEqual(point.name, "P")

    def test_set_name_valid_str(self):
        point = self.getPoint_generic()
        point.name = "P"
        self.assertEqual(point.name, "P")

    def test_set_name_valid_none(self):
        point = self.getPoint_generic()
        point.name = None
        self.assertEqual(point.name, None)

    def test_set_name_invalidType(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.name = 1

    # identifier

    def test_identifier_get_none(self):
        point = self.getPoint_generic()
        self.assertIsNone(point.identifier)

    def test_identifier_generated_type(self):
        point = self.getPoint_generic()
        point.getIdentifier()
        self.assertIsInstance(point.identifier, str)

    def test_identifier_consistency(self):
        point = self.getPoint_generic()
        point.getIdentifier()
        # get: twice to test consistency
        self.assertEqual(point.identifier, point.identifier)

    def test_identifier_cannot_set(self):
        # identifier is a read-only property
        point = self.getPoint_generic()
        with self.assertRaises(FontPartsError):
            point.identifier = "ABC"

    def test_getIdentifer_no_contour(self):
        point, _ = self.objectGenerator("point")
        with self.assertRaises(FontPartsError):
            point.getIdentifier()

    def test_getIdentifer_consistency(self):
        point = self.getPoint_generic()
        point.getIdentifier()
        self.assertEqual(point.identifier, point.getIdentifier())

    # ----
    # Hash
    # ----

    def test_hash_object_self(self):
        point_one = self.getPoint_generic()
        self.assertEqual(hash(point_one), hash(point_one))

    def test_hash_object_other(self):
        point_one = self.getPoint_generic()
        point_two = self.getPoint_generic()
        self.assertNotEqual(hash(point_one), hash(point_two))

    def test_hash_object_self_variable_assignment(self):
        point_one = self.getPoint_generic()
        a = point_one
        self.assertEqual(hash(point_one), hash(a))

    def test_hash_object_other_variable_assignment(self):
        point_one = self.getPoint_generic()
        point_two = self.getPoint_generic()
        a = point_one
        self.assertNotEqual(hash(point_two), hash(a))

    def test_is_hashable(self):
        point_one = self.getPoint_generic()
        self.assertTrue(isinstance(point_one, collections.abc.Hashable))

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        point_one = self.getPoint_generic()
        self.assertEqual(point_one, point_one)

    def test_object_not_equal_other(self):
        point_one = self.getPoint_generic()
        point_two = self.getPoint_generic()
        self.assertNotEqual(point_one, point_two)

    def test_object_equal_self_variable_assignment(self):
        point_one = self.getPoint_generic()
        a = point_one
        self.assertEqual(point_one, a)

    def test_object_not_equal_other_variable_assignment(self):
        point_one = self.getPoint_generic()
        point_two = self.getPoint_generic()
        a = point_one
        self.assertNotEqual(point_two, a)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        point = self.getPoint_generic()
        try:
            point.selected = False
        except NotImplementedError:
            return
        point.selected = True
        self.assertEqual(point.selected, True)

    def test_selected_false(self):
        point = self.getPoint_generic()
        try:
            point.selected = False
        except NotImplementedError:
            return
        self.assertEqual(point.selected, False)

    # ----
    # Copy
    # ----

    def test_copy_seperate_objects(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertIsNot(point, copied)

    def test_copy_different_contour(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertIsNot(point.contour, copied.contour)

    def test_copy_none_contour(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertEqual(copied.contour, None)

    def test_copy_same_type(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertEqual(point.type, copied.type)

    def test_copy_same_smooth(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertEqual(point.smooth, copied.smooth)

    def test_copy_same_x(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertEqual(point.x, copied.x)

    def test_copy_same_y(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertEqual(point.y, copied.y)

    def test_copy_same_name(self):
        point = self.getPoint_generic()
        copied = point.copy()
        self.assertEqual(point.name, copied.name)

    def test_copy_same_identifier_None(self):
        point = self.getPoint_generic()
        point._setIdentifier(None)
        copied = point.copy()
        self.assertEqual(point.identifier, copied.identifier)

    def test_copy_different_identifier(self):
        point = self.getPoint_generic()
        point.getIdentifier()
        copied = point.copy()
        self.assertNotEqual(point.identifier, copied.identifier)

    def test_copy_generated_identifier_different(self):
        otherContour, _ = self.objectGenerator("contour")
        point = self.getPoint_generic()
        copied = point.copy()
        copied.contour = otherContour
        point.getIdentifier()
        copied.getIdentifier()
        self.assertNotEqual(point.identifier, copied.identifier)

    def test_copyData_type(self):
        point = self.getPoint_generic()
        pointOther, _ = self.objectGenerator("point")
        pointOther.copyData(point)
        self.assertEqual(point.type, pointOther.type)

    def test_copyData_smooth(self):
        point = self.getPoint_generic()
        pointOther, _ = self.objectGenerator("point")
        pointOther.copyData(point)
        self.assertEqual(point.smooth, pointOther.smooth)

    def test_copyData_x(self):
        point = self.getPoint_generic()
        pointOther, _ = self.objectGenerator("point")
        pointOther.copyData(point)
        self.assertEqual(point.x, pointOther.x)

    def test_copyData_y(self):
        point = self.getPoint_generic()
        pointOther, _ = self.objectGenerator("point")
        pointOther.copyData(point)
        self.assertEqual(point.y, pointOther.y)

    def test_copyData_name(self):
        point = self.getPoint_generic()
        point.name = "P"
        pointOther, _ = self.objectGenerator("point")
        pointOther.copyData(point)
        self.assertEqual(point.name, pointOther.name)

    def test_copyData_different_identifier(self):
        point = self.getPoint_generic()
        point.getIdentifier()
        pointOther, _ = self.objectGenerator("point")
        pointOther.copyData(point)
        self.assertNotEqual(point.identifier, pointOther.identifier)

    # --------------
    # Transformation
    # --------------

    def test_transformBy_valid_no_origin(self):
        point = self.getPoint_generic()
        point.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual((point.x, point.y), (199.0, 608.0))

    def test_transformBy_valid_origin(self):
        point = self.getPoint_generic()
        point.transformBy((2, 0, 0, 2, 0, 0), origin=(1, 2))
        self.assertEqual((point.x, point.y), (201.0, 402.0))

    def test_transformBy_invalid_one_string_value(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.transformBy((1, 0, 0, 1, 0, "0"))

    def test_transformBy_invalid_all_string_values(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.transformBy("1, 0, 0, 1, 0, 0")

    def test_transformBy_invalid_int_value(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.transformBy(123)

    # moveBy

    def test_moveBy_valid(self):
        point = self.getPoint_generic()
        point.moveBy((-1, 2))
        self.assertEqual((point.x, point.y), (100.0, 204.0))

    def test_moveBy_invalid_one_string_value(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.moveBy((-1, "2"))

    def test_moveBy_invalid_all_strings_value(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.moveBy("-1, 2")

    def test_moveBy_invalid_int_value(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.moveBy(1)

    # scaleBy

    def test_scaleBy_valid_one_value_no_origin(self):
        point = self.getPoint_generic()
        point.scaleBy(-2)
        self.assertEqual((point.x, point.y), (-202.0, -404.0))

    def test_scaleBy_valid_two_values_no_origin(self):
        point = self.getPoint_generic()
        point.scaleBy((-2, 3))
        self.assertEqual((point.x, point.y), (-202.0, 606.0))

    def test_scaleBy_valid_two_values_origin(self):
        point = self.getPoint_generic()
        point.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual((point.x, point.y), (-199.0, 602.0))

    def test_scaleBy_invalid_one_string_value(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.scaleBy((-1, "2"))

    def test_scaleBy_invalid_two_string_values(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.scaleBy("-1, 2")

    def test_scaleBy_invalid_tuple_too_many_values(self):
        point = self.getPoint_generic()
        with self.assertRaises(ValueError):
            point.scaleBy((-1, 2, -3))

    # rotateBy

    def test_rotateBy_valid_no_origin(self):
        point = self.getPoint_generic()
        point.rotateBy(45)
        self.assertEqual([(round(point.x, 3)), (round(point.y, 3))], [-71.418, 214.253])

    def test_rotateBy_valid_origin(self):
        point = self.getPoint_generic()
        point.rotateBy(45, origin=(1, 2))
        self.assertEqual([(round(point.x, 3)), (round(point.y, 3))], [-69.711, 214.132])

    def test_rotateBy_invalid_string_value(self):
        point = self.getPoint_generic()
        with self.assertRaises(TypeError):
            point.rotateBy("45")

    def test_rotateBy_invalid_too_large_value_positive(self):
        point = self.getPoint_generic()
        with self.assertRaises(ValueError):
            point.rotateBy(361)

    def test_rotateBy_invalid_too_large_value_negative(self):
        point = self.getPoint_generic()
        with self.assertRaises(ValueError):
            point.rotateBy(-361)

    # skewBy

    def test_skewBy_valid_no_origin_one_value(self):
        point = self.getPoint_generic()
        point.skewBy(100)
        self.assertEqual([(round(point.x, 3)), (round(point.y, 3))], [-1044.599, 202.0])

    def test_skewBy_valid_no_origin_two_values(self):
        point = self.getPoint_generic()
        point.skewBy((100, 200))
        self.assertEqual(
            [(round(point.x, 3)), (round(point.y, 3))], [-1044.599, 238.761]
        )

    def test_skewBy_valid_origin_one_value(self):
        point = self.getPoint_generic()
        point.skewBy(100, origin=(1, 2))
        self.assertEqual([(round(point.x, 3)), (round(point.y, 3))], [-1033.256, 202.0])

    def test_skewBy_valid_origin_two_values(self):
        point = self.getPoint_generic()
        point.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(
            [(round(point.x, 3)), (round(point.y, 3))], [-1033.256, 238.397]
        )

    # -------------
    # Normalization
    # -------------

    # round

    def getPoint_floatXY(self):
        point, _ = self.objectGenerator("point")
        point.x = 101.3
        point.y = 202.6
        return point

    def test_roundX(self):
        point = self.getPoint_floatXY()
        point.round()
        self.assertEqual(point.x, 101)

    def test_roundY(self):
        point = self.getPoint_floatXY()
        point.round()
        self.assertEqual(point.y, 203)



import unittest
import collections


class TestLayer(unittest.TestCase):
    # ------
    # Glyphs
    # ------

    def getLayer_glyphs(self):
        layer, _ = self.objectGenerator("layer")
        for name in "ABCD":
            layer.newGlyph(name)
        return layer

    def test_len(self):
        layer = self.getLayer_glyphs()
        self.assertEqual(len(layer), 4)

    def _testInsertGlyph(self, setGlyphName=True):
        layer, _ = self.objectGenerator("layer")
        glyph, _ = self.objectGenerator("glyph")
        pen = glyph.getPen()
        pen.moveTo((100, 0))
        pen.lineTo((100, 500))
        pen.lineTo((500, 500))
        pen.closePath()
        glyph.width = 600
        if setGlyphName:
            glyph.name = "test"
        layer["test"] = glyph
        self.assertTrue("test" in layer)
        self.assertEqual(layer["test"].bounds, glyph.bounds)

    def test_set_glyph(self):
        self._testInsertGlyph(setGlyphName=True)

    def test_set_glyph_with_name_None(self):
        self._testInsertGlyph(setGlyphName=False)

    def test_get_glyph_in_font(self):
        layer = self.getLayer_glyphs()
        self.assertEqual(layer["A"].name, "A")

    def test_get_glyph_not_in_font(self):
        layer = self.getLayer_glyphs()
        with self.assertRaises(KeyError):
            layer["E"]

    # ----
    # Hash
    # ----

    def test_hash_object_self(self):
        layer_one = self.getLayer_glyphs()
        self.assertEqual(hash(layer_one), hash(layer_one))

    def test_hash_object_other(self):
        layer_one = self.getLayer_glyphs()
        layer_two = self.getLayer_glyphs()
        self.assertNotEqual(hash(layer_one), hash(layer_two))

    def test_hash_object_self_variable_assignment(self):
        layer_one = self.getLayer_glyphs()
        a = layer_one
        self.assertEqual(hash(layer_one), hash(a))

    def test_hash_object_other_variable_assignment(self):
        layer_one = self.getLayer_glyphs()
        layer_two = self.getLayer_glyphs()
        a = layer_one
        self.assertNotEqual(hash(layer_two), hash(a))

    def test_is_hashable(self):
        layer_one = self.getLayer_glyphs()
        self.assertTrue(isinstance(layer_one, collections.abc.Hashable))

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        layer_one = self.getLayer_glyphs()
        self.assertEqual(layer_one, layer_one)

    def test_object_not_equal_other(self):
        layer_one = self.getLayer_glyphs()
        layer_two = self.getLayer_glyphs()
        self.assertNotEqual(layer_one, layer_two)

    def test_object_equal_self_variable_assignment(self):
        layer_one = self.getLayer_glyphs()
        a = layer_one
        self.assertEqual(layer_one, a)

    def test_object_not_equal_self_variable_assignment(self):
        layer_one = self.getLayer_glyphs()
        layer_two = self.getLayer_glyphs()
        a = layer_one
        self.assertNotEqual(layer_two, a)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        layer.selected = True
        self.assertEqual(layer.selected, True)

    def test_selected_false(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        self.assertEqual(layer.selected, False)

    # Glyphs

    def test_selectedGlyphs_default(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        self.assertEqual(layer.selectedGlyphs, ())

    def test_selectedGlyphs_setSubObject(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        glyph1 = layer["A"]
        glyph2 = layer["B"]
        glyph1.selected = True
        glyph2.selected = True
        self.assertEqual(
            tuple(sorted(layer.selectedGlyphs, key=lambda glyph: glyph.name)),
            (glyph1, glyph2),
        )

    def test_selectedGlyphs_setFilledList(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        glyph3 = layer["C"]
        glyph4 = layer["D"]
        layer.selectedGlyphs = [glyph3, glyph4]
        self.assertEqual(
            tuple(sorted(layer.selectedGlyphs, key=lambda glyph: glyph.name)),
            (glyph3, glyph4),
        )

    def test_selectedGlyphs_setEmptyList(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        glyph1 = layer["A"]
        glyph1.selected = True
        layer.selectedGlyphs = []
        self.assertEqual(layer.selectedGlyphs, ())

    # Glyph Names

    def test_selectedGlyphNames_default(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        self.assertEqual(layer.selectedGlyphs, ())

    def test_selectedGlyphNames_setSubObject(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        glyph1 = layer["A"]
        glyph2 = layer["B"]
        glyph1.selected = True
        glyph2.selected = True
        self.assertEqual(tuple(sorted(layer.selectedGlyphNames)), ("A", "B"))

    def test_selectedGlyphNames_setFilledList(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        layer.selectedGlyphNames = ["C", "D"]
        self.assertEqual(tuple(sorted(layer.selectedGlyphNames)), ("C", "D"))

    def test_selectedGlyphNames_setEmptyList(self):
        layer = self.getLayer_glyphs()
        try:
            layer.selected = False
        except NotImplementedError:
            return
        glyph1 = layer["A"]
        glyph1.selected = True
        layer.selectedGlyphNames = []
        self.assertEqual(layer.selectedGlyphNames, ())



from fontParts.test import testEnvironment
from fontParts.fontshell.font import RFont
from fontParts.fontshell.info import RInfo
from fontParts.fontshell.groups import RGroups
from fontParts.fontshell.kerning import RKerning
from fontParts.fontshell.features import RFeatures
from fontParts.fontshell.layer import RLayer
from fontParts.fontshell.glyph import RGlyph
from fontParts.fontshell.contour import RContour
from fontParts.fontshell.segment import RSegment
from fontParts.fontshell.bPoint import RBPoint
from fontParts.fontshell.point import RPoint
from fontParts.fontshell.anchor import RAnchor
from fontParts.fontshell.component import RComponent
from fontParts.fontshell.image import RImage
from fontParts.fontshell.lib import RLib
from fontParts.fontshell.guideline import RGuideline


# defcon does not have prebuilt support for
# selection states, so we simulate selection
# behavior with a small subclasses for testing
# purposes only.


def _get_selected(self):
    if isinstance(self, FSTestSegment):
        for point in self.points:
            if point.selected:
                return True
        return False
    elif isinstance(self, FSTestBPoint):
        point = self._point.naked()
        return point.name == "selected"
    elif isinstance(self, FSTestPoint):
        return self.name == "selected"
    else:
        if not hasattr(self.naked(), "_testSelected"):
            return False
        return self.naked()._testSelected


def _set_selected(self, value):
    if isinstance(self, FSTestSegment):
        for point in self.points:
            point.selected = value
    elif isinstance(self, FSTestBPoint):
        point = self._point.naked()
        if value:
            point.name = "selected"
        else:
            point.name = None
    elif isinstance(self, FSTestPoint):
        if value:
            self.name = "selected"
        else:
            self.name = None
    else:
        self.naked()._testSelected = value


class FSTestPoint(RPoint):
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestBPoint(RBPoint):
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestSegment(RSegment):
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestGuideline(RGuideline):
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestImage(RImage):
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestAnchor(RAnchor):
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestComponent(RComponent):
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestContour(RContour):
    segmentClass = FSTestSegment
    bPointClass = FSTestBPoint
    pointClass = FSTestPoint
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestGlyph(RGlyph):
    contourClass = FSTestContour
    componentClass = FSTestComponent
    anchorClass = FSTestAnchor
    guidelineClass = FSTestGuideline
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestLayer(RLayer):
    glyphClass = FSTestGlyph
    _get_selected = _get_selected
    _set_selected = _set_selected


class FSTestFont(RFont):
    layerClass = FSTestLayer
    guidelineClass = FSTestGuideline
    _get_selected = _get_selected
    _set_selected = _set_selected


classMapping = dict(
    font=FSTestFont,
    info=RInfo,
    groups=RGroups,
    kerning=RKerning,
    features=RFeatures,
    layer=FSTestLayer,
    glyph=FSTestGlyph,
    contour=FSTestContour,
    segment=FSTestSegment,
    bPoint=FSTestBPoint,
    point=FSTestPoint,
    anchor=FSTestAnchor,
    component=FSTestComponent,
    image=FSTestImage,
    lib=RLib,
    guideline=FSTestGuideline,
)


def fontshellObjectGenerator(cls):
    unrequested = []
    obj = classMapping[cls]()
    return obj, unrequested


if __name__ == "__main__":
    import sys

    if {"-v", "--verbose"}.intersection(sys.argv):
        verbosity = 2
    else:
        verbosity = 1
    testEnvironment(fontshellObjectGenerator, verbosity=verbosity)



import unittest
import collections
from fontParts.base import FontPartsError


class TestContour(unittest.TestCase):
    def getContour_bounds(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 100), "line")
        contour.appendPoint((100, 100), "line")
        contour.appendPoint((100, 0), "line")
        return contour

    # ----
    # Repr
    # ----

    def test_reprContents_noGlyph_noID(self):
        contour = self.getContour_bounds()
        value = contour._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_noGlyph_ID(self):
        contour = self.getContour_bounds()
        contour.getIdentifier()
        value = contour._reprContents()
        self.assertIsInstance(value, list)
        idFound = False
        for i in value:
            self.assertIsInstance(i, str)
            if i == f"identifier='{contour.identifier!r}'":
                idFound = True
        self.assertTrue(idFound)

    def test_reprContents_Glyph_ID(self):
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        contour.glyph = glyph
        contour.getIdentifier()
        value = contour._reprContents()
        self.assertIsInstance(value, list)
        idFound = False
        glyphFound = False
        for i in value:
            self.assertIsInstance(i, str)
            if i == f"identifier='{contour.identifier!r}'":
                idFound = True
            if i == "in glyph":
                glyphFound = True
        self.assertTrue(idFound)
        self.assertTrue(glyphFound)

    def test_reprContents_Glyph_noID(self):
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        contour.glyph = glyph
        value = contour._reprContents()
        self.assertIsInstance(value, list)
        glyphFound = False
        for i in value:
            self.assertIsInstance(i, str)
            if i == "in glyph":
                glyphFound = True
        self.assertTrue(glyphFound)

    # ----
    # Copy
    # ----

    def test_copyData(self):
        contour = self.getContour_bounds()
        contourOther, _ = self.objectGenerator("contour")
        contourOther.copyData(contour)
        self.assertEqual(contour.bounds, contourOther.bounds)

    # -------
    # Parents
    # -------

    def test_parent_glyph_set_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        contour.glyph = glyph
        self.assertEqual(glyph, contour.glyph)

    def test_parent_glyph_set_glyph_None(self):
        contour = self.getContour_bounds()
        contour.glyph = None
        self.assertEqual(None, contour.glyph)

    def test_parent_glyph_set_already_set(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph2, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        contour.glyph = glyph
        self.assertEqual(glyph, contour.glyph)
        with self.assertRaises(AssertionError):
            contour.glyph = glyph2

    def test_parent_glyph_get_none(self):
        contour = self.getContour_bounds()
        self.assertEqual(None, contour.glyph)

    def test_parent_glyph_get(self):
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        contour = glyph.appendContour(contour)
        self.assertEqual(glyph, contour.glyph)

    def test_parent_font_set(self):
        font, _ = self.objectGenerator("font")
        contour = self.getContour_bounds()
        with self.assertRaises(FontPartsError):
            contour.font = font

    def test_parent_font_get_none(self):
        contour = self.getContour_bounds()
        self.assertEqual(None, contour.font)

    def test_parent_font_get(self):
        font, _ = self.objectGenerator("font")
        layer, _ = self.objectGenerator("layer")
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        layer.font = font
        glyph.layer = layer
        contour = glyph.appendContour(contour)
        self.assertEqual(font, contour.font)

    def test_parent_layer_set(self):
        layer, _ = self.objectGenerator("layer")
        contour = self.getContour_bounds()
        with self.assertRaises(FontPartsError):
            contour.layer = layer

    def test_parent_layer_get_none(self):
        contour = self.getContour_bounds()
        self.assertEqual(None, contour.layer)

    def test_parent_layer_get(self):
        layer, _ = self.objectGenerator("layer")
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        glyph.layer = layer
        contour = glyph.appendContour(contour)
        self.assertEqual(layer, contour.layer)

    # ----
    # Pens
    # ----

    def test_drawPoints_oldProtocol(self):
        contour = self.getContour_bounds()

        class OldProtocolPen:
            def __init__(self):
                self.calls = []

            def beginPath(self):
                self.calls.append(("beginPath",))

            def addPoint(self, pt, segmentType=None, smooth=False, name=None):
                self.calls.append(("addPoint", pt, segmentType, smooth, name))

            def endPath(self):
                self.calls.append(("endPath",))

        pen = OldProtocolPen()
        contour.drawPoints(pen)

        self.assertEqual(len(pen.calls), 6)
        self.assertEqual(pen.calls[0], ("beginPath",))
        self.assertEqual(pen.calls[-1], ("endPath",))

    # -------------
    # Normalization
    # -------------

    def test_round(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 99.5), "line")
        contour.appendPoint((99.6, 100.4), "line")
        contour.appendPoint((100, 0), "line")
        contour.round()
        result = self.getContour_bounds()
        self.assertEqual(contour.bounds, result.bounds)

    # -------------
    # Interpolation
    # -------------

    def test_isCompatible_same_contours(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        compatible, report = contour1.isCompatible(contour2)
        self.assertTrue(compatible)
        self.assertFalse(report.fatal)
        self.assertFalse(report.warning)

    def test_isCompatible_different_segment_count(self):
        contour1 = self.getContour_bounds()
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((0, 0), "line")
        contour2.appendPoint((50, 100), "line")
        contour2.appendPoint((100, 0), "line")
        compatible, report = contour1.isCompatible(contour2)
        self.assertFalse(compatible)
        self.assertTrue(report.fatal)
        self.assertTrue(report.segmentCountDifference)

    def test_isCompatible_different_open_closed(self):
        contour1 = self.getContour_bounds()
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((0, 0), "move")
        contour2.appendPoint((0, 100), "line")
        contour2.appendPoint((100, 100), "line")
        contour2.appendPoint((100, 0), "line")
        compatible, report = contour1.isCompatible(contour2)
        self.assertFalse(compatible)
        self.assertTrue(report.openDifference)
        self.assertTrue(report.fatal)

    def test_isCompatible_different_direction(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        contour2.reverse()
        compatible, report = contour1.isCompatible(contour2)
        self.assertTrue(compatible)
        self.assertFalse(report.fatal)
        self.assertFalse(report.warning)
        self.assertTrue(report.directionDifference)

    def test_isCompatible_incompatible_segments(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        contour2[0].type = "qcurve"
        compatible, report = contour1.isCompatible(contour2)
        self.assertFalse(compatible)
        self.assertTrue(report.fatal)
        self.assertTrue(len(report.segments) > 0)

    def test_isCompatible_multiple_issues(self):
        contour1 = self.getContour_bounds()
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((0, 0), "move")
        contour2.appendPoint((0, 100), "line")
        contour2.appendPoint((100, 100), "line")
        contour2.appendPoint((100, 0), "line")
        contour2[1].type = "qcurve"
        compatible, report = contour1.isCompatible(contour2)
        self.assertFalse(compatible)
        self.assertTrue(report.openDifference)
        self.assertTrue(report.fatal)

    # ---------
    # Direction
    # ---------

    def test_set_clockwise(self):
        contour = self.getContour_bounds()
        contour.clockwise = True
        self.assertTrue(contour.clockwise)
        contour.clockwise = False
        self.assertFalse(contour.clockwise)

    # ------------------------
    # Point and Contour Inside
    # ------------------------

    def test_pointInside(self):
        contour = self.getContour_bounds()
        self.assertTrue(contour.pointInside((50, 50)))
        self.assertFalse(contour.pointInside((200, 200)))

    def test_contourInside(self):
        outerContour = self.getContour_bounds()
        innerContour, _ = self.objectGenerator("contour")
        innerContour.appendPoint((25, 25), "line")
        innerContour.appendPoint((25, 75), "line")
        innerContour.appendPoint((75, 75), "line")
        innerContour.appendPoint((75, 25), "line")
        self.assertTrue(outerContour.contourInside(innerContour))
        self.assertFalse(innerContour.contourInside(outerContour))

    # -----
    # Index
    # -----

    def test_index(self):
        glyph, _ = self.objectGenerator("glyph")
        contour1 = glyph.appendContour(self.getContour_bounds())
        contour2 = glyph.appendContour(self.getContour_bounds())
        contour3 = glyph.appendContour(self.getContour_bounds())
        self.assertEqual(contour1.index, 0)
        self.assertEqual(contour2.index, 1)
        self.assertEqual(contour3.index, 2)

    def test_set_index(self):
        glyph, _ = self.objectGenerator("glyph")
        contour1 = glyph.appendContour(self.getContour_bounds())
        contour2 = glyph.appendContour(self.getContour_bounds())
        contour3 = glyph.appendContour(self.getContour_bounds())

        contour1.index = 2
        self.assertEqual(contour1.index, 2)
        self.assertEqual(contour2.index, 0)
        self.assertEqual(contour3.index, 1)

        contour1.index = 1
        self.assertEqual(contour1.index, 1)
        self.assertEqual(contour2.index, 0)
        self.assertEqual(contour3.index, 2)

        contour1.index = 3
        self.assertEqual(contour1.index, 2)

    def test_set_index_orphan_contour(self):
        contour = self.getContour_bounds()
        with self.assertRaises(FontPartsError):
            contour.index = 1

    def test_set_index_negative_value(self):
        glyph, _ = self.objectGenerator("glyph")
        contour1 = glyph.appendContour(self.getContour_bounds())
        contour2 = glyph.appendContour(self.getContour_bounds())
        contour1.index = -1
        self.assertEqual(contour1.index, 0)
        self.assertEqual(contour2.index, 1)
        contour1.index = -11
        self.assertEqual(contour1.index, 0)
        self.assertEqual(contour2.index, 1)

    def test_set_index_none(self):
        glyph, _ = self.objectGenerator("glyph")
        contour1 = glyph.appendContour(self.getContour_bounds())
        contour2 = glyph.appendContour(self.getContour_bounds())
        contour1.index = None
        self.assertEqual(contour1.index, 0)
        self.assertEqual(contour2.index, 1)

    # --------------
    # Identification
    # --------------

    def test_get_index_no_glyph(self):
        contour = self.getContour_bounds()
        self.assertEqual(contour.index, None)

    def test_get_index_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        c1 = glyph.appendContour(contour)
        self.assertEqual(c1.index, 0)
        c2 = glyph.appendContour(contour)
        self.assertEqual(c2.index, 1)

    # ------
    # Bounds
    # ------

    def getContour_boundsExtrema(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 100), "line")
        contour.appendPoint((50, 100), "line")
        contour.appendPoint((117, 100), "offcurve")
        contour.appendPoint((117, 0), "offcurve")
        contour.appendPoint((50, 0), "curve")
        return contour

    def test_bounds_get(self):
        contour = self.getContour_bounds()
        self.assertEqual(contour.bounds, (0, 0, 100, 100))

    def test_bounds_set_float(self):
        contour = self.getContour_bounds()
        contour.moveBy((0.5, -0.5))
        self.assertEqual(contour.bounds, (0.5, -0.5, 100.5, 99.5))

    def test_bounds_point_not_at_extrema(self):
        contour = self.getContour_bounds()
        contour = self.getContour_boundsExtrema()
        bounds = tuple(int(round(i)) for i in contour.bounds)
        self.assertEqual(bounds, (0, 0, 100, 100))

    def test_invalid_bounds_set(self):
        contour = self.getContour_bounds()
        with self.assertRaises(FontPartsError):
            contour.bounds = (1, 2, 3, 4)

    def test_empty_bounds(self):
        contour, _ = self.objectGenerator("contour")
        self.assertIsNone(contour.bounds)

    # ----
    # Area
    # ----

    def test_area(self):
        contour = self.getContour_bounds()
        self.assertEqual(contour.area, 10000)

    def test_empty_area(self):
        contour, _ = self.objectGenerator("contour")
        self.assertIsNone(contour.area)

    # ----
    # Hash
    # ----

    def test_hash_object_self(self):
        contour_one = self.getContour_bounds()
        self.assertEqual(hash(contour_one), hash(contour_one))

    def test_hash_object_other(self):
        contour_one = self.getContour_bounds()
        contour_two = self.getContour_bounds()
        self.assertNotEqual(hash(contour_one), hash(contour_two))

    def test_hash_object_self_variable_assignment(self):
        contour_one = self.getContour_bounds()
        a = contour_one
        self.assertEqual(hash(contour_one), hash(a))

    def test_hash_object_other_variable_assignment(self):
        contour_one = self.getContour_bounds()
        contour_two = self.getContour_bounds()
        a = contour_one
        self.assertNotEqual(hash(contour_two), hash(a))

    def test_is_hashable(self):
        contour_one = self.getContour_bounds()
        self.assertTrue(isinstance(contour_one, collections.abc.Hashable))

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        contour_one = self.getContour_bounds()
        self.assertEqual(contour_one, contour_one)

    def test_object_not_equal_self(self):
        contour_one = self.getContour_bounds()
        contour_two = self.getContour_bounds()
        self.assertNotEqual(contour_one, contour_two)

    def test_object_equal_self_variable_assignment(self):
        contour_one = self.getContour_bounds()
        a = contour_one
        a.moveBy((0.5, -0.5))
        self.assertEqual(contour_one, a)

    def test_object_not_equal_self_variable_assignment(self):
        contour_one = self.getContour_bounds()
        contour_two = self.getContour_bounds()
        a = contour_one
        self.assertNotEqual(contour_two, a)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        contour = self.getContour_bounds()
        try:
            contour.selected = False
        except NotImplementedError:
            return
        contour.selected = True
        self.assertEqual(contour.selected, True)

    def test_selected_false(self):
        contour = self.getContour_bounds()
        try:
            contour.selected = False
        except NotImplementedError:
            return
        self.assertEqual(contour.selected, False)

    def test_selectedSegments_default(self):
        contour = self.getContour_bounds()
        segment1 = contour.segments[0]
        try:
            segment1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(contour.selectedSegments, ())

    def test_selectedSegments_setSubObject(self):
        contour = self.getContour_bounds()
        segment1 = contour.segments[0]
        segment2 = contour.segments[1]
        try:
            segment1.selected = False
        except NotImplementedError:
            return
        segment2.selected = True
        self.assertEqual(contour.selectedSegments == (segment2,), True)

    def test_selectedSegments_setFilledList(self):
        contour = self.getContour_bounds()
        segment1 = contour.segments[0]
        segment2 = contour.segments[1]
        try:
            segment1.selected = False
        except NotImplementedError:
            return
        contour.selectedSegments = [segment1, segment2]
        self.assertEqual(contour.selectedSegments, (segment1, segment2))

    def test_selectedSegments_setEmptyList(self):
        contour = self.getContour_bounds()
        segment1 = contour.segments[0]
        try:
            segment1.selected = True
        except NotImplementedError:
            return
        contour.selectedSegments = []
        self.assertEqual(contour.selectedSegments, ())

    def test_selectedPoints_default(self):
        contour = self.getContour_bounds()
        point1 = contour.points[0]
        try:
            point1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(contour.selectedPoints, ())

    def test_selectedPoints_setSubObject(self):
        contour = self.getContour_bounds()
        point1 = contour.points[0]
        point2 = contour.points[1]
        try:
            point1.selected = False
        except NotImplementedError:
            return
        point2.selected = True
        self.assertEqual(contour.selectedPoints, (point2,))

    def test_selectedPoints_setFilledList(self):
        contour = self.getContour_bounds()
        point1 = contour.points[0]
        point2 = contour.points[1]
        try:
            point1.selected = False
        except NotImplementedError:
            return
        contour.selectedPoints = [point1, point2]
        self.assertEqual(contour.selectedPoints, (point1, point2))

    def test_selectedPoints_setEmptyList(self):
        contour = self.getContour_bounds()
        point1 = contour.points[0]
        try:
            point1.selected = True
        except NotImplementedError:
            return
        contour.selectedPoints = []
        self.assertEqual(contour.selectedPoints, ())

    def test_selectedBPoints_default(self):
        contour = self.getContour_bounds()
        bPoint1 = contour.bPoints[0]
        try:
            bPoint1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(contour.selectedBPoints, ())

    def test_selectedBPoints_setSubObject(self):
        contour = self.getContour_bounds()
        bPoint1 = contour.bPoints[0]
        bPoint2 = contour.bPoints[1]
        try:
            bPoint1.selected = False
        except NotImplementedError:
            return
        bPoint2.selected = True
        self.assertEqual(contour.selectedBPoints, (bPoint2,))

    def test_selectedBPoints_setFilledList(self):
        contour = self.getContour_bounds()
        bPoint1 = contour.bPoints[0]
        bPoint2 = contour.bPoints[1]
        try:
            bPoint1.selected = False
        except NotImplementedError:
            return
        contour.selectedBPoints = [bPoint1, bPoint2]
        self.assertEqual(contour.selectedBPoints, (bPoint1, bPoint2))

    def test_selectedBPoints_setEmptyList(self):
        contour = self.getContour_bounds()
        bPoint1 = contour.bPoints[0]
        try:
            bPoint1.selected = True
        except NotImplementedError:
            return
        contour.selectedBPoints = []
        self.assertEqual(contour.selectedBPoints, ())

    # --------
    # Segments
    # --------

    def test_segments_offcurves_end(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((84, 0), "curve")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 28), "offcurve")
        contour.appendPoint((10, 64), "offcurve")
        contour.appendPoint((46, 64), "curve")
        contour.appendPoint((76, 64), "offcurve")
        contour.appendPoint((84, 28), "offcurve")

        segments = contour.segments
        self.assertEqual(
            [segment.type for segment in segments], ["line", "curve", "curve"]
        )

    def test_segments_offcurves_begin_end(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((84, 28), "offcurve")
        contour.appendPoint((84, 0), "curve")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 28), "offcurve")
        contour.appendPoint((10, 64), "offcurve")
        contour.appendPoint((46, 64), "curve")
        contour.appendPoint((76, 64), "offcurve")

        segments = contour.segments
        self.assertEqual(
            [segment.type for segment in segments], ["line", "curve", "curve"]
        )

    def test_segments_offcurves_begin(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((76, 64), "offcurve")
        contour.appendPoint((84, 28), "offcurve")
        contour.appendPoint((84, 0), "curve")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 28), "offcurve")
        contour.appendPoint((10, 64), "offcurve")
        contour.appendPoint((46, 64), "curve")

        segments = contour.segments
        self.assertEqual(
            [segment.type for segment in segments], ["line", "curve", "curve"]
        )

    def test_segments_offcurves_middle(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((46, 64), "curve")
        contour.appendPoint((76, 64), "offcurve")
        contour.appendPoint((84, 28), "offcurve")
        contour.appendPoint((84, 0), "curve")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 28), "offcurve")
        contour.appendPoint((10, 64), "offcurve")

        segments = contour.segments
        self.assertEqual(
            [segment.type for segment in segments], ["curve", "line", "curve"]
        )

    def test_segments_start_move_offcurves_end(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((84, 28), "move")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 28), "offcurve")
        contour.appendPoint((10, 64), "offcurve")
        contour.appendPoint((46, 64), "curve")
        contour.appendPoint((76, 64), "offcurve")

        segments = contour.segments
        self.assertEqual(
            [segment.type for segment in segments], ["move", "line", "curve"]
        )

    def test_segments_empty(self):
        contour, _ = self.objectGenerator("contour")
        segments = contour.segments
        self.assertEqual(segments, ())

    def test_segments_segmentClass_none(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        conntourClass = type(contour)
        originalClass = conntourClass.segmentClass
        try:
            conntourClass.segmentClass = None
            with self.assertRaises(TypeError):
                contour.segments
        finally:
            conntourClass.segmentClass = originalClass

    def test_segment_insert_open(self):
        # at index 0
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((2, 2), "line")
        contour.appendPoint((3, 3), "line")
        contour.insertSegment(0, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (1, 1), (2, 2), (3, 3)],
        )
        # at index 1
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((2, 2), "line")
        contour.appendPoint((3, 3), "line")
        contour.insertSegment(1, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (2, 2), (1, 1), (3, 3)],
        )

    def test_segment_insert_curve_open(self):
        # at index 0
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((2, 2), "offcurve")
        contour.appendPoint((3, 3), "offcurve")
        contour.appendPoint((4, 4), "curve")
        contour.appendPoint((5, 5), "line")
        contour.insertSegment(0, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5)],
        )
        # at index 1
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((2, 2), "offcurve")
        contour.appendPoint((3, 3), "offcurve")
        contour.appendPoint((4, 4), "curve")
        contour.appendPoint((5, 5), "line")
        contour.insertSegment(1, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (2, 2), (3, 3), (4, 4), (1, 1), (5, 5)],
        )

    def test_segment_insert_closed(self):
        # at index 0
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((2, 2), "line")
        contour.appendPoint((3, 3), "line")
        contour.insertSegment(0, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (1, 1), (2, 2), (3, 3)],
        )
        # at index 1
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((2, 2), "line")
        contour.appendPoint((3, 3), "line")
        contour.insertSegment(1, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (2, 2), (1, 1), (3, 3)],
        )

    def test_segment_insert_curve_closed(self):
        # at index 0
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((2, 2), "offcurve")
        contour.appendPoint((3, 3), "offcurve")
        contour.appendPoint((4, 4), "curve")
        contour.appendPoint((5, 5), "line")
        contour.insertSegment(0, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5)],
        )
        # at index 1
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((2, 2), "offcurve")
        contour.appendPoint((3, 3), "offcurve")
        contour.appendPoint((4, 4), "curve")
        contour.appendPoint((5, 5), "line")
        contour.insertSegment(1, "line", [(1, 1)])
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(0, 0), (2, 2), (3, 3), (4, 4), (1, 1), (5, 5)],
        )

    def test_setStartSegment_index(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((50, 0), "curve")
        contour.appendPoint((75, 0), "offcurve")
        contour.appendPoint((100, 25), "offcurve")
        contour.appendPoint((100, 50), "curve")
        contour.appendPoint((100, 75), "offcurve")
        contour.appendPoint((75, 100), "offcurve")
        contour.appendPoint((50, 100), "curve")
        contour.appendPoint((25, 100), "offcurve")
        contour.appendPoint((0, 75), "offcurve")
        contour.appendPoint((0, 50), "curve")
        contour.appendPoint((0, 25), "offcurve")
        contour.appendPoint((25, 0), "offcurve")
        contour.setStartSegment(1)
        self.assertEqual(contour.points[0].type, "curve")
        self.assertEqual((contour.points[0].x, contour.points[0].y), (100, 50))

    def test_setStartSegment_segment(self):
        contour = self.getContour_bounds()
        targetSegment = contour.segments[1]
        contour.setStartSegment(targetSegment)
        self.assertEqual(contour.points[0].position, (0, 100))

    def test_setStartSegment_index_out_of_range(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((100, 0), "line")
        with self.assertRaises(ValueError):
            contour.setStartSegment(5)

    def test_setStartSegment_open_contour(self):
        contour, _ = self.objectGenerator("contour")
        with self.assertRaises(FontPartsError):
            contour.setStartSegment(1)

    def test_iterSegment(self):
        contour = self.getContour_bounds()
        iterator = iter(contour)
        self.assertIsInstance(iterator, collections.abc.Iterator)
        segments = list(iterator)
        self.assertEqual(len(segments), len(contour.segments))
        self.assertEqual([segment.type for segment in segments], ["line"] * 4)

    def test_appendSegment_points(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((10, 0), "line")
        initialLength = len(contour)
        points = [(50, 50)]
        contour.appendSegment(type="line", points=points)
        appendedPoints = [p.position for p in contour.segments[-1].points]
        self.assertEqual(len(contour), initialLength + 1)
        self.assertEqual(appendedPoints, points)

    def test_appendSegment_segment(self):
        contour1, _ = self.objectGenerator("contour")
        contour1.appendPoint((0, 0), "move")
        contour1.appendPoint((10, 0), "line")
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "line")
        segment = contour2.segments[0]
        points = [p.position for p in segment]
        contour1.appendSegment(segment=segment)
        appendedPoints = [p.position for p in contour1.segments[-1].points]
        self.assertEqual(len(contour1), initialLength + 1)
        self.assertEqual(appendedPoints, points)

    def test_appendSegment_overrides(self):
        contour1, _ = self.objectGenerator("contour")
        contour1.appendPoint((0, 0), "move")
        contour1.appendPoint((10, 0), "line")
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50))
        segment = contour2.segments[0]
        overridePoints = [(120, 120)]
        contour1.appendSegment(type="curve", segment=segment, points=overridePoints)
        appendedPoints = [p.position for p in contour1.segments[-1].points]
        self.assertEqual(len(contour1), initialLength + 1)
        self.assertEqual(appendedPoints, overridePoints)

    def test_appendSegment_raises(self):
        contour = self.getContour_bounds()
        with self.assertRaises(TypeError):
            contour.appendSegment(type=None, points=[(50, 50)])
        with self.assertRaises(TypeError):
            contour.appendSegment(type="line", points=None)

    def test_insertSegment_segment(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "line")
        contour2.appendPoint((60, 60), "offcurve")
        segment = contour2.segments[0]
        contour1.insertSegment(0, segment=segment)
        insertedPoints = [p.position for p in contour1.segments[0].points]
        originalPoints = [p.position for p in segment.points]
        self.assertEqual(len(contour1), initialLength + 1)
        self.assertEqual(insertedPoints, originalPoints)
        self.assertEqual(contour1.segments[0].type, "line")

    def test_insertSegment_overrides(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50))
        segment = contour2.segments[0]
        overridePoints = [(120, 120)]
        contour1.insertSegment(0, type="curve", segment=segment, points=overridePoints)
        insertedPoints = [p.position for p in contour1.segments[0].points]
        self.assertEqual(len(contour1), initialLength + 1)
        self.assertEqual(insertedPoints, overridePoints)
        self.assertEqual(contour1.segments[0].type, "curve")

    def test_insertSegment_raises(self):
        contour = self.getContour_bounds()
        with self.assertRaises(TypeError):
            contour.insertSegment(None, type="line", points=[(50, 50)])
        with self.assertRaises(TypeError):
            contour.insertSegment(0, type=None, points=[(50, 50)])
        with self.assertRaises(TypeError):
            contour.insertSegment(0, type="line", points=None)

    def test_removeSegment_index(self):
        contour = self.getContour_bounds()
        initialLength = len(contour)
        contour.removeSegment(0)
        self.assertEqual(len(contour), initialLength - 1)

    def test_removeSegment_index_out_of_range(self):
        contour = self.getContour_bounds()
        with self.assertRaises(ValueError):
            contour.removeSegment(10)

    def test_removeSegment_segment(self):
        contour = self.getContour_bounds()
        initialLength = len(contour)
        segment = contour.segments[0]
        contour.removeSegment(segment)
        self.assertEqual(len(contour), initialLength - 1)

    # -------
    # bPoints
    # -------

    def test_bPoints_bPointClass_none(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0))
        contourClass = type(contour)
        originalClass = contourClass.bPointClass
        try:
            contourClass.bPointClass = None
            with self.assertRaises(TypeError):
                contour.bPoints
        finally:
            contourClass.bPointClass = originalClass

    def test_appendBPoint_values(self):
        contour = self.getContour_bounds()
        initialLength = len(contour.bPoints)
        contour.appendBPoint(type="corner", anchor=(0, 0))
        self.assertEqual(len(contour.bPoints), initialLength + 1)
        self.assertEqual(contour.bPoints[-1].type, "corner")
        self.assertEqual(contour.bPoints[-1].anchor, (0, 0))

    def test_appendBPoint_bPoint(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "line")
        bPoint = contour2.bPoints[0]
        contour1.appendBPoint(bPoint=bPoint)
        self.assertEqual(len(contour1.bPoints), initialLength + 1)
        self.assertEqual(contour1.bPoints[-1].type, "corner")
        self.assertEqual(contour1.bPoints[-1].anchor, (50, 50))

    def test_appendBPoint_overrides(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "line")
        bPoint = contour2.bPoints[0]
        contour1.appendBPoint(
            type="curve", anchor=(0, 0), bcpIn=(10, 10), bcpOut=(20, 20), bPoint=bPoint
        )
        self.assertEqual(len(contour1.bPoints), initialLength + 1)
        self.assertEqual(contour1.bPoints[-1].type, "curve")
        self.assertEqual(contour1.bPoints[-1].anchor, (0, 0))
        self.assertEqual(contour1.bPoints[-1].bcpIn, (10, 10))
        self.assertEqual(contour1.bPoints[-1].bcpOut, (20, 20))

    def test_insertBPoint_values(self):
        contour = self.getContour_bounds()
        initialLength = len(contour.bPoints)
        contour.insertBPoint(0, type="curve", anchor=(50, 50))
        self.assertEqual(len(contour.bPoints), initialLength + 1)
        self.assertEqual(contour.bPoints[1].anchor, (50, 50))

    def test_insertBPoint_bPoint(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "line")
        bPoint = contour2.bPoints[0]
        contour1.insertBPoint(0, bPoint=bPoint)
        self.assertEqual(len(contour1.bPoints), initialLength + 1)
        self.assertEqual(contour1.bPoints[1].type, "corner")
        self.assertEqual(contour1.bPoints[1].anchor, (50, 50))

    def test_insertBPoint_overrides(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "line")
        bPoint = contour2.bPoints[0]
        contour1.insertBPoint(
            0,
            type="curve",
            anchor=(0, 0),
            bcpIn=(10, 10),
            bcpOut=(20, 20),
            bPoint=bPoint,
        )
        self.assertEqual(len(contour1.bPoints), initialLength + 1)
        self.assertEqual(contour1.bPoints[1].type, "curve")
        self.assertEqual(contour1.bPoints[1].anchor, (0, 0))
        self.assertEqual(contour1.bPoints[1].bcpIn, (10, 10))
        self.assertEqual(contour1.bPoints[1].bcpOut, (20, 20))

    def test_insertBPoint_raises(self):
        contour = self.getContour_bounds()
        with self.assertRaises(TypeError):
            contour.insertBPoint(None, type="corner", anchor=(50, 50))
        with self.assertRaises(TypeError):
            contour.insertBPoint(0, type=None, anchor=(50, 50))
        with self.assertRaises(TypeError):
            contour.insertBPoint(0, type="corner", anchor=None)

    def test_removeBPoint_index(self):
        contour = self.getContour_bounds()
        initialLength = len(contour.bPoints)
        contour.removeBPoint(0)
        self.assertEqual(len(contour.bPoints), initialLength - 1)
        self.assertEqual(contour.bPoints[0].anchor, (0, 100))

    def test_removeBPoint_index_out_of_range(self):
        contour = self.getContour_bounds()
        with self.assertRaises(ValueError):
            contour.removeBPoint(10)

    def test_removeBPoint_bPoint(self):
        contour = self.getContour_bounds()
        initialLength = len(contour)
        bPoint = contour.bPoints[0]
        contour.removeBPoint(bPoint)
        self.assertEqual(len(contour.bPoints), initialLength - 1)

    def test_removeBPoint_offcurves(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "curve")
        contour.appendPoint((0, 50), "offcurve")
        contour.appendPoint((50, 100), "offcurve")
        contour.appendPoint((100, 100), "curve")
        contour.appendPoint((150, 100), "offcurve")
        contour.appendPoint((200, 50), "offcurve")
        initialLength = len(contour.bPoints)
        contour.removeBPoint(-2)
        self.assertEqual(len(contour.bPoints), initialLength - 1)

    # ------
    # points
    # ------

    def test_setStartPoint(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((1, 1), "line")
        contour.appendPoint((2, 2), "line")
        contour.appendPoint((3, 3), "line")

        contour.setStartPoint(2)
        self.assertEqual(
            [(point.x, point.y) for point in contour.points],
            [(2, 2), (3, 3), (0, 0), (1, 1)],
        )

    def test_setStartPoint_index_zero(self):
        contour = self.getContour_bounds()
        contour.setStartPoint(0)
        self.assertEqual(contour.points[0].position, (0, 0))

    def test_setStartPoint_index_out_of_range(self):
        contour = self.getContour_bounds()
        with self.assertRaises(ValueError):
            contour.setStartPoint(10)

    def test_setStartPoint_open_contour(self):
        contour, _ = self.objectGenerator("contour")
        with self.assertRaises(FontPartsError):
            contour.setStartPoint(0)

    def test_appendPoint_position(self):
        contour = self.getContour_bounds()
        initialLength = len(contour.points)
        contour.appendPoint(position=(10, 10))
        self.assertEqual(len(contour.bPoints), initialLength + 1)
        self.assertEqual(contour.points[-1].position, (10, 10))
        self.assertEqual(contour.points[-1].type, "line")

    def test_appendPoint_bPoint(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "curve")
        point = contour2.points[0]
        contour1.appendPoint(point=point)
        self.assertEqual(len(contour1.points), initialLength + 1)
        self.assertEqual(contour1.points[-1].position, (50, 50))
        self.assertEqual(contour1.points[-1].type, "curve")

    def test_appendPoint_overrides(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "curve")
        point = contour2.points[0]
        contour1.appendPoint(
            position=(10, 10), type="line", name="test", identifier="test", point=point
        )
        self.assertEqual(len(contour1.points), initialLength + 1)
        self.assertEqual(contour1.points[-1].position, (10, 10))
        self.assertEqual(contour1.points[-1].type, "line")
        self.assertEqual(contour1.points[-1].name, "test")
        self.assertEqual(contour1.points[-1].identifier, "test")

    def test_insertPoint_position(self):
        contour = self.getContour_bounds()
        initialLength = len(contour.points)
        contour.insertPoint(0, position=(10, 10))
        self.assertEqual(len(contour.bPoints), initialLength + 1)
        self.assertEqual(contour.points[0].position, (10, 10))
        self.assertEqual(contour.points[0].type, "line")

    def test_insertPoint_point(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "curve")
        point = contour2.points[0]
        contour1.insertPoint(0, point=point)
        self.assertEqual(len(contour1.points), initialLength + 1)
        self.assertEqual(contour1.points[0].position, (50, 50))
        self.assertEqual(contour1.points[0].type, "curve")

    def test_insertPoint_overrides(self):
        contour1 = self.getContour_bounds()
        initialLength = len(contour1)
        contour2, _ = self.objectGenerator("contour")
        contour2.appendPoint((50, 50), "curve")
        point = contour2.points[0]
        contour1.insertPoint(
            0,
            position=(10, 10),
            type="line",
            name="test",
            identifier="test",
            point=point,
        )
        self.assertEqual(len(contour1.points), initialLength + 1)
        self.assertEqual(contour1.points[0].position, (10, 10))
        self.assertEqual(contour1.points[0].type, "line")
        self.assertEqual(contour1.points[0].name, "test")
        self.assertEqual(contour1.points[0].identifier, "test")

    def test_insertPoint_raises(self):
        contour = self.getContour_bounds()
        with self.assertRaises(TypeError):
            contour.insertPoint(None, type="line", position=(50, 50))
        with self.assertRaises(TypeError):
            contour.insertPoint(0, type="line", position=None)

    def test_removePoint_index_out_of_range(self):
        contour = self.getContour_bounds()
        with self.assertRaises(ValueError):
            contour.removePoint(10)
